"""
Derived module from dmdbase.py for hankel dmd.

Reference:
- H. Arbabi, I. Mezic, Ergodic theory, dynamic mode decomposition, and
computation of spectral properties of the Koopman operator. SIAM Journal on
Applied Dynamical Systems, 2017, 16.4: 2096-2126.
"""

from copy import copy
from numbers import Number

import numpy as np

from .dmd import DMD
from .dmdbase import DMDBase
from .preprocessing.hankel import hankel_preprocessing
from .snapshots import Snapshots


class HankelDMD(DMDBase):
    """
    Hankel Dynamic Mode Decomposition

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means no truncation.
    :param bool exact: flag to compute either exact DMD or projected DMD.
        Default is False.
    :param opt: argument to control the computation of DMD modes amplitudes.
        See :class:`DMDBase`. Default is False.
    :type opt: bool or int
    :param rescale_mode: Scale Atilde as shown in
            10.1016/j.jneumeth.2015.10.010 (section 2.4) before computing its
            eigendecomposition. None means no rescaling, 'auto' means automatic
            rescaling using singular values, otherwise the scaling factors.
    :type rescale_mode: {'auto'} or None or numpy.ndarray
    :param bool forward_backward: If True, the low-rank operator is computed
        like in fbDMD (reference: https://arxiv.org/abs/1507.02264). Default is
        False.
    :param int d: the new order for spatial dimension of the input snapshots.
        Default is 1.
    :param sorted_eigs: Sort eigenvalues (and modes/dynamics accordingly) by
        magnitude if `sorted_eigs='abs'`, by real part (and then by imaginary
        part to break ties) if `sorted_eigs='real'`. Default: False.
    :type sorted_eigs: {'real', 'abs'} or False
    :param reconstruction_method: Method used to reconstruct the snapshots of
        the dynamical system from the multiple versions available due to how
        HankelDMD is conceived. If `'first'` (default) the first version
        available is selected (i.e. the nearest to the 0-th row in the
        augmented matrix). If `'mean'` we compute the element-wise mean. If
        `reconstruction_method` is an array of float values we compute the
        weighted average (for each snapshots) using the given values as weights
        (the number of weights must be equal to `d`).
    :type reconstruction_method: {'first', 'mean'} or array-like
    """

    def __init__(
        self,
        svd_rank=0,
        tlsq_rank=0,
        exact=False,
        opt=False,
        rescale_mode=None,
        forward_backward=False,
        d=1,
        sorted_eigs=False,
        reconstruction_method="first",
        tikhonov_regularization=None,
    ):
        super().__init__(
            svd_rank=svd_rank,
            tlsq_rank=tlsq_rank,
            exact=exact,
            opt=opt,
            rescale_mode=rescale_mode,
            sorted_eigs=sorted_eigs,
        )
        self._d = d
        self._ho_snapshots = None

        if isinstance(reconstruction_method, list):
            if len(reconstruction_method) != d:
                raise ValueError(
                    "The length of the array of weights must be equal to d"
                )
        elif isinstance(reconstruction_method, np.ndarray):
            if (
                reconstruction_method.ndim > 1
                or reconstruction_method.shape[0] != d
            ):
                raise ValueError(
                    "The length of the array of weights must be equal to d"
                )
        self._reconstruction_method = reconstruction_method

        sub_dmd = DMD(
            svd_rank=svd_rank,
            tlsq_rank=tlsq_rank,
            exact=exact,
            opt=opt,
            rescale_mode=rescale_mode,
            forward_backward=forward_backward,
            sorted_eigs=sorted_eigs,
            tikhonov_regularization=tikhonov_regularization,
        )
        self._sub_dmd = hankel_preprocessing(
            sub_dmd, d=d, reconstruction_method=reconstruction_method
        )

    def _hankel_first_occurrence(self, time):
        r"""
        For a given `t` such that there is :math:`k \in \mathbb{N}` such that
        :math:`t = t_0 + k dt`, return the index of the first column in Hankel
        pseudo matrix (see also :func:`_pseudo_hankel_matrix`) which contains
        the snapshot corresponding to `t`.

        :param time: The time corresponding to the requested snapshot.
        :return: The index of the first appeareance of `time` in the columns of
            Hankel pseudo matrix.
        :rtype: int
        """
        assert isinstance(time, Number) or np.asarray(time).ndim == 1
        return max(
            0,
            (time - self.original_time["t0"]) // self.dmd_time["dt"]
            - (self.original_time["t0"] + self._d - 1),
        )

    def _update_sub_dmd_time(self):
        """
        Update the time dictionaries (`dmd_time` and `original_time`) of
        the auxiliary DMD instance `HankelDMD._sub_dmd` after an update of the
        time dictionaries of the time dictionaries of this instance of the
        higher level instance of `HankelDMD`.
        """
        self._sub_dmd.dmd_time["t0"] = self._hankel_first_occurrence(
            self.dmd_time["t0"]
        )
        self._sub_dmd.dmd_time["tend"] = self._hankel_first_occurrence(
            self.dmd_time["tend"]
        )

    @property
    def reconstructed_data(self):
        self._update_sub_dmd_time()

        # we want to return only the requested timesteps
        time_index = min(
            self._d - 1,
            int(
                (self.dmd_time["t0"] - self.original_time["t0"])
                // self.dmd_time["dt"]
            ),
        )
        return self._sub_dmd.reconstructed_data[
            :, time_index : time_index + len(self.dmd_timesteps)
        ]

    @property
    def modes(self):
        return self._sub_dmd.modes

    @property
    def eigs(self):
        return self._sub_dmd.eigs

    @property
    def amplitudes(self):
        return self._sub_dmd.amplitudes

    @property
    def operator(self):
        return self._sub_dmd.operator

    @property
    def svd_rank(self):
        return self._sub_dmd.svd_rank

    @property
    def ho_snapshots(self):
        """
        Get the time-delay data matrix.

        :return: the matrix that contains the time-delayed data.
        :rtype: numpy.ndarray
        """
        return self._sub_dmd.snapshots

    @property
    def modes_activation_bitmask(self):
        return self._sub_dmd.modes_activation_bitmask

    @modes_activation_bitmask.setter
    def modes_activation_bitmask(self, value):
        self._sub_dmd.modes_activation_bitmask = value

    def __getitem__(self, key):
        # The implementation was asking for problems...
        raise ValueError("This operation is not allowed for HankelDMD")

    def fit(self, X):
        """
        Compute the Dynamic Modes Decomposition to the input data.

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        """
        self._reset()

        self._snapshots_holder = Snapshots(X)

        n_samples = self.snapshots.shape[-1]
        if n_samples < self._d:
            msg = """The number of snapshots provided is not enough for d={}.
Expected at least d."""
            raise ValueError(msg.format(self._d))

        self._sub_dmd.fit(X)

        # Default timesteps
        self._set_initial_time_dictionary(
            {"t0": 0, "tend": n_samples - 1, "dt": 1}
        )

        return self



"""
Variable Projection for DMD. Reformulation of original paper
(https://epubs.siam.org/doi/abs/10.1137/M1124176) s.t. sparse matrix computation
is substiuted by outer products. Further the optimization is
reformulated s.t. SciPy's nonlinear least squares optimizer
can handle "complex" parameters.

Default optimizer arguments:
    OPT_DEF_ARGS =
        {"method": "trf",
        "tr_solver": "exact",
        "loss": "linear",
        "x_scale": "jac",
        "gtol": 1e-8,
        "xtol": 1e-8,
        "ftol": 1e-8}
"""

import warnings
from types import MappingProxyType
from typing import Any, Dict, Tuple, Union

import numpy as np
from scipy.linalg import qr
from scipy.optimize import OptimizeResult, least_squares

from .dmd import DMDBase
from .dmdoperator import DMDOperator
from .snapshots import Snapshots
from .utils import compute_svd

OPT_DEF_ARGS = MappingProxyType(
    {
        "method": "trf",
        "tr_solver": "exact",
        "loss": "linear",
        "x_scale": "jac",
        "gtol": 1e-8,
        "xtol": 1e-8,
        "ftol": 1e-8,
    }
)


def _compute_dmd_ev(
    x_current: np.ndarray,
    x_next: np.ndarray,
    rank: Union[float, int] = 0,
) -> np.ndarray:
    r"""
    Compute DMD eigenvalues.

    :param x_current: Observables :math:`\boldsymbol{X}`
    :type x_current: np.ndarray
    :param x_next: Observables :math:`\boldsymbol{X}'`
    :type x_current: np.ndarray
    :param rank: Desired rank. If rank :math:`r = 0`, the optimal rank is
        determined automatically. If rank is a float s.t. :math: `0 < r < 1`,
        the cumulative energy of the singular values is used
        to determine the optimal rank. If rank is an integer and :math:`r > 0`,
        the desired rank is used iff possible. Defaults to 0.
    :type rank: Union[float, int], optional
    :return: Diagonal matrix of eigenvalues
        :math:`\boldsymbol{\Lambda}` as 1d array.
    :rtype: np.ndarray
    """

    u_x, sigma_x, v_x = compute_svd(x_current, rank)

    # columns of v need to be multiplicated with inverse sigma
    sigma_inv_approx = np.reciprocal(sigma_x)

    a_approx = np.linalg.multi_dot(
        [u_x.conj().T, x_next, sigma_inv_approx[None] * v_x]
    )

    return np.linalg.eigvals(a_approx)


class _OptimizeHelper:  # pylint: disable=too-few-public-methods
    """
    Helper Class to store intermediate results during the optimization.
    """

    __slots__ = ["phi", "phi_inv", "u_svd", "s_inv", "v_svd", "b_matrix", "rho"]

    def __init__(self, l_in: int, m_in: int, n_in: int):
        self.phi = np.empty((m_in, l_in), dtype=np.complex128)
        self.u_svd = np.empty((m_in, l_in), dtype=np.complex128)
        self.s_inv = np.empty((l_in,), dtype=np.complex128)
        self.v_svd = np.empty((l_in, l_in), dtype=np.complex128)
        self.b_matrix = np.empty((l_in, n_in), dtype=np.complex128)
        self.rho = np.empty((m_in, n_in), dtype=np.complex128)


def _compute_dmd_rho(
    alphas: np.ndarray,
    time: np.ndarray,
    data: np.ndarray,
    opthelper: _OptimizeHelper,
) -> np.ndarray:
    r"""
    Compute the real residual vector :math:`\boldsymbol{\rho}` for
    Levenberg-Marquardt update.

    :param alphas: DMD eigenvalues to optimize,
        where :math:`\alpha \in \mathbb{C}^l`,
        but here :math:`\alpha \in \mathbb{R}^{2l}`,
        since optimizer cannot deal with complex numbers.
    :type alphas: np.ndarray
    :param time: 1D time array.
    :type time: np.ndarray
    :param data: data :math:`\boldsymbol{Y} \n C^{m \times n}`.
        For DMD computation we set :math:`\boldsymbol{Y} = \boldsymbol{X}^T`.
    :type data: np.ndarray
    :param opthelper: Optimization helper to speed up computations
        mainly for Jacobian.
    :type opthelper: _OptimizeHelper
    :return: 1d residual vector for Levenberg-Marquardt update
        :math:`\boldsymbol{\rho} \in \mathbb{R}^{2mn}`.
    :rtype: np.ndarray
    """

    _alphas = np.zeros((alphas.shape[-1] // 2,), dtype=np.complex128)
    _alphas.real = alphas[: alphas.shape[-1] // 2]
    _alphas.imag = alphas[alphas.shape[-1] // 2 :]

    phi = np.exp(np.outer(time, _alphas))
    u_phi, s_phi, v_phi_t = np.linalg.svd(phi, full_matrices=False)
    idx = np.where(s_phi.real != 0.0)[0]
    s_phi_inv = np.zeros_like(s_phi)
    s_phi_inv[idx] = np.reciprocal(s_phi[idx])

    rho = data - np.linalg.multi_dot([u_phi, u_phi.conj().T, data])
    rho_flat = np.ravel(rho)
    rho_out = np.zeros((2 * rho_flat.shape[-1],), dtype=np.float64)
    rho_out[: rho_flat.shape[-1]] = rho_flat.real
    rho_out[rho_flat.shape[-1] :] = rho_flat.imag

    opthelper.phi = phi
    opthelper.u_svd = u_phi
    opthelper.s_inv = s_phi_inv
    opthelper.v_svd = v_phi_t.conj().T
    opthelper.rho = rho
    opthelper.b_matrix = np.linalg.multi_dot(
        [
            opthelper.v_svd * s_phi_inv[None],
            opthelper.u_svd.conj().T,
            data,
        ]
    )
    return rho_out


def _compute_dmd_jac(
    alphas: np.ndarray,
    time: np.ndarray,
    data: np.ndarray,
    opthelper: _OptimizeHelper,
) -> np.ndarray:
    r"""
    Compute the real Jacobian.
    SciPy's nonlinear least squares optimizer requires real entities.
    Therefore, complex and real parts are split.

    :param alphas: DMD eigenvalues to optimize,
        where :math:`\alpha \in \mathbb{C}^l`,
        but here :math:`\alpha \in \mathbb{R}^{2l}` since optimizer cannot
        deal with complex numbers.
    :type alphas: np.ndarray
    :param time: 1D time array.
    :type time: np.ndarray
    :param data: data :math: `\boldsymbol{Y} \n C^{m \times n}`.
        For DMD computation we set :math:`\boldsymbol{Y} = \boldsymbol{X}^T`.
    :type data: np.ndarray
    :param opthelper: Optimization helper to speed up computations
        mainly for Jacobian. The entities are computed in `_compute_dmd_rho`.
    :type opthelper: _OptimizeHelper
    :return: Jacobian :math:`\boldsymbol{J} \in \mathbb{R}^{mn \times 2l}`.
    :rtype: np.ndarray
    """

    _alphas = np.zeros((alphas.shape[-1] // 2,), dtype=np.complex128)
    _alphas.real = alphas[: alphas.shape[-1] // 2]
    _alphas.imag = alphas[alphas.shape[-1] // 2 :]
    jac_out = np.zeros((2 * np.prod(data.shape), alphas.shape[-1]))

    for j in range(_alphas.shape[-1]):
        d_phi_j = time * opthelper.phi[:, j]
        outer = np.outer(d_phi_j, opthelper.b_matrix[j])
        a_j = outer - np.linalg.multi_dot(
            [opthelper.u_svd, opthelper.u_svd.conj().T, outer]
        )
        g_j = np.linalg.multi_dot(
            [
                opthelper.u_svd * opthelper.s_inv[None],
                np.outer(
                    opthelper.v_svd[j].conj(), d_phi_j.conj() @ opthelper.rho
                ),
            ]
        )
        # Compute the jacobian J_mat_j = - (A_j + G_j).
        jac = -a_j - g_j
        jac_flat = np.ravel(jac)

        # construct the overall jacobian for optimized
        # J_real = |Re{J} -Im{J}|
        #          |Im{J}  Re{J}|

        # construct real part for optimization
        jac_out[: jac_out.shape[0] // 2, j] = jac_flat.real
        jac_out[jac_out.shape[0] // 2 :, j] = jac_flat.imag

        # construct imaginary part for optimization
        jac_out[: jac_out.shape[0] // 2, _alphas.shape[-1] + j] = -jac_flat.imag
        jac_out[jac_out.shape[0] // 2 :, _alphas.shape[-1] + j] = jac_flat.real

    return jac_out


def _varpro_preprocessing(
    data: np.ndarray,
    time: np.ndarray,
    rank: Union[float, int] = 0.0,
    use_proj: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    r"""
    Preprocess data for Variable Projection: Calculate
    :math:`\boldsymbol{Y}, \boldsymbol{Z}` for trapezoidal derivative
    approximation. If desired input data is projected to
    low-dimensional space.

    :param data: data matrix s.t. :math:`X \n C^{n \times m}`.
    :type data: np.ndarray
    :param time: 1d array of timestamps.
    :type time: np.ndarray
    :param optargs: Arguments for 'least_squares' optimizer.
    :type optargs: Dict[str, Any]
    :param rank: Desired rank. If rank :math:`r = 0`, the optimal rank is
        determined automatically. If rank is a float s.t. :math:`0 < r < 1`,
        the cumulative energy of the singular values is used
        to determine the optimal rank. If rank is an integer
        and :math:`r > 0`, the desired rank is used iff possible.
        Defaults to 0.
    :type rank: Union[float, int], optional
    :param use_proj: Perform variable projection in
        low dimensional space if `use_proj=True`, else in the original space.
        Defaults to True.
    :type use_proj: bool, optional
    :return: Derivative :math:`\boldsymbol{Y},\boldsymbol{Z}`, (projected) data,
         rank reduced projection matrix :math:`\boldsymbol{U}_r`.
    :rtype: Tuple[np.ndarray,
                  np.ndarray,
                  np.ndarray,
                  np.ndarray]
    """

    u_r, s_r, v_r = compute_svd(data, rank)
    data_out = v_r.conj().T * s_r[:, None] if use_proj else data

    # trapezoidal derivative approximation
    y_out = (data_out[:, :-1] + data_out[:, 1:]) / 2.0
    dt_in = time[1:] - time[:-1]
    z_out = (data_out[:, 1:] - data_out[:, :-1]) / dt_in[None]

    return z_out, y_out, data_out, u_r


def _compute_dmd_varpro(
    alphas_init: np.ndarray,
    time: np.ndarray,
    data: np.ndarray,
    opthelper: _OptimizeHelper,
    **optargs,
) -> OptimizeResult:
    r"""
    Compute Variable Projection (VarPro) for DMD using SciPy's
    nonlinear least squares optimizer.

    :type alphas_init: np.ndarray
    :param time: 1d time array.
    :type time: np.ndarray
    :param data: data :math:`\boldsymbol{Y} \n C^{m \times n}`.
        For DMD computation we set :math:`\boldsymbol{Y} = \boldsymbol{X}^T`.
    :type data: np.ndarray
    :param opthelper: Optimization helper to speed up computations
        mainly for Jacobian. The entities are computed in `_compute_dmd_rho`
        and are used in `_compute_dmd_jac`.
    :type opthelper: _OptimizeHelper
    :return: Optimization result.
    :rtype: OptimizeResult
    """

    return least_squares(
        _compute_dmd_rho,
        alphas_init,
        _compute_dmd_jac,
        **optargs,
        args=[time, data, opthelper],
    )


def select_best_samples_fast(data: np.ndarray, comp: float = 0.9) -> np.ndarray:
    r"""
    Select library samples using QR decomposition with column pivoting.

    :param data: Data matrix :math:`\boldsymbol{X} \in \mathbb{C}^{n \times m}`.
    :type data: np.ndarray
    :param comp: Library compression :math:`c`, where :math:`0 < c < 1`.
        The best fitting :math:`\lfloor \left(1 - c\right)m\rfloor` samples
        are selected. Defaults to 0.9.
    :type comp: float, optional
    :raises ValueError: ValueError is raised if data matrix is not a
        2d array.
    :raises ValueError: ValueError is raised of compression is not in required
        interval (:math:`0 < r < 1`).
    :return: Indices of selected samples as 1d array.
    :rtype: np.ndarray
    """

    if len(data.shape) != 2:
        raise ValueError("Expected 2D array!")

    if not 0 < comp < 1:
        raise ValueError("Compression must be in (0, 1)]")

    n_samples = int(data.shape[-1] * (1.0 - comp))
    pcolumn = qr(data, mode="economic", pivoting=True)[-1]

    return pcolumn[:n_samples]


def compute_varprodmd_any(  # pylint: disable=unused-variable
    data: np.ndarray,
    time: np.ndarray,
    optargs: Dict[str, Any],
    rank: Union[float, int] = 0.0,
    use_proj: bool = True,
    compression: float = 0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, OptimizeResult]:
    r"""
    Compute DMD given arbitary timesteps.

    :param data: data matrix s.t. :math:`X \n C^{n \times m}`.
    :type data: np.ndarray
    :param time: 1d array of timestamps.
    :type time: np.ndarray
    :param optargs: Arguments for 'least_squares' optimizer.
    :type optargs: Dict[str, Any]
    :param rank: Desired rank. If rank :math:`r = 0`, the optimal rank is
        determined automatically. If rank is a float s.t. :math:`0 < r < 1`,
        the cumulative energy of the singular values is used
        to determine the optimal rank. If rank is an integer
        and :math:`r > 0`, the desired rank is used iff possible.
        Defaults to 0.
    :type rank: Union[float, int], optional
    :param use_proj: Perform variable projection in
        low dimensional space if `use_proj=True`, else in the original space.
        Defaults to True.
    :type use_proj: bool, optional
    :param compression: If libary compression :math:`c = 0`,
        all samples are used. If :math:`0 < c < 1`, the best
        fitting :math:`\lfloor \left(1 - c\right)m\rfloor` samples
        are selected.
    :type compression: float, optional
    :raises ValueError: ValueError is raised if data matrix is not a
        2d array.
    :raises ValueError: ValueError is raised if time is not a
        1d array.
    :return: DMD modes :math:`\boldsymbol{\Phi}`, continuous DMD eigenvalues
        :math: `\boldsymbol{\Omega}` as 1d array,
        DMD eigenfunctions or amplitudes :math:`\boldsymbol{\varphi}`,
        optimization results of SciPy's nonlinear least squares optimizer.
    :rtype: Tuple[np.ndarray,
                  np.ndarray,
                  np.ndarray,
                  np.ndarray,
                  OptimizeResult]
    """

    if len(data.shape) != 2:
        raise ValueError("data needs to be 2D array")

    if len(time.shape) != 1:
        raise ValueError("time needs to be a 1D array")

    #  y_in, z_in, data_in, u_r
    res = _varpro_preprocessing(data, time, rank, use_proj)
    omegas = _compute_dmd_ev(res[0], res[1], res[-1].shape[-1])

    if compression > 0:
        indices = select_best_samples_fast(res[2], compression)

        if indices.size <= 1:
            indices = np.arange(res[2])

    else:
        indices = np.arange(res[2].shape[-1])

    if res[2].shape[-1] < omegas.shape[-1]:
        warnings.warn(
            " ".join(
                (
                    "Attempting to solve underdetermined system.",
                    "Decrease desired rank or compression!",
                )
            )
        )

    opthelper = _OptimizeHelper(res[-1].shape[-1], *res[2].shape)
    opt = _compute_dmd_varpro(
        np.concatenate([omegas.real, omegas.imag]),
        time[indices],
        res[2][:, indices].T,
        opthelper,
        **optargs,
    )
    omegas.real = opt.x[: opt.x.shape[-1] // 2]
    omegas.imag = opt.x[opt.x.shape[-1] // 2 :]
    xi = res[-1] @ opthelper.b_matrix.T if use_proj else opthelper.b_matrix.T
    eigenf = np.linalg.norm(xi, axis=0)
    return xi / eigenf[None], omegas, eigenf, indices, opt


def varprodmd_predict(
    phi: np.ndarray,
    omegas: np.ndarray,
    eigenf: np.ndarray,
    time: np.ndarray,
) -> np.ndarray:
    r"""
    Perform DMD prediction using computed modes, continuous eigenvalues
    and eigenfunctions/amplitudes.

    :param phi: DMD modes
        :math:`\boldsymbol{\Phi} \in \mathbb{C}^{n \times \left(m-1\right)}`.
    :type phi: np.ndarray
    :param omegas: Continuous diagonal matrix of eigenvalues
        :math:`\boldsymbol{\Omega} \in
            \mathbb{C}^{\left(m-1\right) \times \left(m-1\right)}`
        as 1d array.
    :type omegas: np.ndarray
    :param eigenf: Eigenfunctions or amplitudes
        :math:`\boldsymbol{\varphi} \in \mathbb{C}^{m - 1}`.
    :type eigenf: np.ndarray
    :param time: 1d array of timestamps.
    :type time: np.ndarray
    :return: Reconstructed/predicted state :math:`\hat{\boldsymbol{X}}`.
    :rtype: np.ndarray
    """

    return phi @ (np.exp(np.outer(omegas, time)) * eigenf[:, None])


class VarProOperator(DMDOperator):
    """
    Variable Projection Operator for DMD.
    """

    def __init__(
        self,
        svd_rank: Union[float, int],
        exact: bool,
        sorted_eigs: Union[bool, str],
        compression: float,
        optargs: Dict[str, Any],
    ):
        r"""
        VarProOperator constructor.

        :param svd_rank: Desired SVD rank.
            If rank :math:`r = 0`, the optimal rank is
            determined automatically. If rank is a float s.t.
            :math:`0 < r < 1`, the cumulative energy
            of the singular values is used to determine
            the optimal rank.
            If rank is an integer and :math:`r > 0`,
            the desired rank is used iff possible.
        :type svd_rank: Union[float, int]
        :param exact: Perform computations in original state space
            if `exact=True`, else perform optimization
            in projected (low dimensional) space.
        :type exact: bool
        :param sorted_eigs: Sort eigenvalues.
            If `sorted_eigs=True`, the variance of the absolute values
            of the complex eigenvalues
            :math:`\sqrt{\omega_i \cdot \bar{\omega}_i}`,
            the variance absolute values of the real parts
            :math:`\left|\Re\{{\omega_i}\}\right|`
            and the variance of the absolute values
            of the imaginary parts :math:`\left|\Im\{{\omega_i}\}\right|`
            is computed.
            The eigenvalues are then sorted according
            to the highest variance (from highest to lowest).
            If `sorted_eigs=False`, no sorting is performed.
            If the parameter is a string and set to sorted_eigs='auto',
            the eigenvalues are sorted accoring to the variances
            of previous mentioned quantities.
            If `sorted_eigs='real'` the eigenvalues are sorted
            w.r.t. the absolute values of the real parts
            of the eigenvalues (from highest to lowest).
            If `sorted_eigs='imag'` the eigenvalues are sorted
            w.r.t. the absolute values of the imaginary parts
            of the eigenvalues (from highest to lowest).
            If `sorted_eigs='abs'` the eigenvalues are sorted
            w.r.t. the magnitude of the eigenvalues
            :math:`\left(\sqrt{\omega_i \cdot \bar{\omega}_i}\right)`
            (from highest to lowest).
        :type sorted_eigs: Union[bool, str]
        :param compression: If libary compression :math:`c = 0`,
            all samples are used. If :math:`0 < c < 1`, the best
            fitting :math:`\lfloor \left(1 - c\right)m\rfloor` samples
            are selected.
        :type compression: float
        :param optargs: Arguments for 'least_squares' optimizer.
            Use `OPT_DEF_ARGS` as starting point.
        :type optargs: Dict[str, Any]
        """

        super().__init__(svd_rank, exact, False, None, sorted_eigs, False)
        self._sorted_eigs = sorted_eigs
        self._svd_rank = svd_rank
        self._exact = exact
        self._optargs: Dict[str, Any] = optargs
        self._compression: float = compression
        self._modes: np.ndarray = None
        self._eigenvalues: np.ndarray = None

    def compute_operator(
        self, X: np.ndarray, Y: np.ndarray
    ) -> Tuple[np.ndarray, OptimizeResult, np.ndarray]:
        r"""
        Perform Variable Projection for DMD using SciPy's
        nonlinear least squares optimizer.

        :param X: Measurement :math:`\boldsymbol{X} \in \mathbb{C}^{n \times m}`
        :type X: np.ndarray
        :param Y: 1d array of timestamps where individual
            measurements :math:`\boldsymbol{x}_i \in \mathbb{C}^n`
            where taken.
        :type Y: np.ndarray
        :raises ValueError: If `sorted_eigs` from constructor
            was set to an invalid string.
        :return: Eigenfunctions/amplitudes :math:`\boldsymbol{\varphi}^{m-1}`,
            OptimizeResult from SciPy's optimizer
            (optimal parameters and statistics),
            indices of selected measurements. If no compression (:math:`c = 0`)
            is used, all indices are returned, else the indices of the selected
            samples are used.
        :rtype: Tuple[np.ndarray, OptimizeResult, np.ndarray]
        """
        (
            self._modes,
            self._eigenvalues,
            eigenf,
            indices,
            opt,
        ) = compute_varprodmd_any(
            X,
            Y,
            self._optargs,
            self._svd_rank,
            not self._exact,
            self._compression,
        )

        # overwrite for lazy sorting
        if isinstance(self._sorted_eigs, bool):
            if self._sorted_eigs:
                self._sorted_eigs = "auto"

        if isinstance(self._sorted_eigs, str):
            if self._sorted_eigs == "auto":
                eigs_real = self._eigenvalues.real
                eigs_imag = self._eigenvalues.imag
                _eigs_abs = np.abs(self._eigenvalues)
                var_real = np.var(eigs_real)
                var_imag = np.var(eigs_imag)
                var_abs = np.var(_eigs_abs)
                array = np.array([var_real, var_imag, var_abs])
                eigs_abs = (eigs_real, eigs_imag, _eigs_abs)[np.argmax(array)]

            elif self._sorted_eigs == "real":
                eigs_abs = np.abs(self._eigenvalues.real)

            elif self._sorted_eigs == "imag":
                eigs_abs = np.abs(self._eigenvalues.imag)

            elif self._sorted_eigs == "abs":
                eigs_abs = np.abs(self._eigenvalues)
            else:
                raise ValueError(f"{self._sorted_eigs} not supported!")

            idx = np.argsort(eigs_abs)[::-1]  # sort from biggest to smallest
            self._eigenvalues = self._eigenvalues[idx]
            self._modes = self._modes[:, idx]
            eigenf = eigenf[idx]

        return eigenf, opt, indices


class VarProDMD(DMDBase):
    """
    Variable Projection for DMD using SciPy's
    nonlinear least squares solver. The original problem
    is reformulated s.t. complex residuals and Jacobians, which are used
    by the Levenberg-Marquardt algorithm, are transormed to real numbers.
    Further simplifications (outer products) avoids using sparse matrices.
    """

    def __init__(  # pylint: disable=super-init-not-called
        self,
        svd_rank: Union[float, int] = 0,
        exact: bool = False,
        sorted_eigs: Union[bool, str] = False,
        compression: float = 0.0,
        optargs: Dict[str, Any] = None,
    ):
        r"""
        VarProDMD constructor.

        :param svd_rank: Desired SVD rank.
            If rank :math:`r = 0`, the optimal rank is
            determined automatically. If rank is a float s.t. :math:`0 < r < 1`,
            the cumulative energy of the singular values is used
            to determine the optimal rank.
            If rank is an integer and :math:`r > 0`,
            the desired rank is used iff possible. Defaults to 0.
        :type svd_rank: Union[float, int], optional
        :param exact: Perform variable projection in
            low dimensional space if `exact=False`.
            Else the optimization is performed
            in the original space.
            Defaults to False.
        :type exact: bool, optional
        :param sorted_eigs: Sort eigenvalues.
            If `sorted_eigs=True`, the variance of the absolute values
            of the complex eigenvalues
            :math:`\left(\sqrt{\omega_i \cdot \bar{\omega}_i}\right)`,
            the variance absolute values of the real parts
            :math:`\left|\Re\{{\omega_i}\}\right|`
            and the variance of the absolute values of the imaginary parts
            :math:`\left|\Im\{{\omega_i}\}\right|` is computed.
            The eigenvalues are then sorted according
            to the highest variance (from highest to lowest).
            If `sorted_eigs=False`, no sorting is performed.
            If the parameter is a string and set to sorted_eigs='auto',
            the eigenvalues are sorted accoring to the variances
            of previous mentioned quantities.
            If `sorted_eigs='real'` the eigenvalues are sorted
            w.r.t. the absolute values of the real parts
            of the eigenvalues (from highest to lowest).
            If `sorted_eigs='imag'` the eigenvalues are sorted
            w.r.t. the absolute values of the imaginary parts
            of the eigenvalues (from highest to lowest).
            If `sorted_eigs='abs'` the eigenvalues are sorted
            w.r.t. the magnitude of the eigenvalues
            :math:`\left(\sqrt{\omega_i \cdot \bar{\omega}_i}\right)`
            (from highest to lowest).
            Defaults to False.
        :type sorted_eigs: Union[bool, str], optional
        :param compression: If libary compression :math:`c = 0`,
            all samples are used. If :math:`0 < c < 1`, the best
            fitting :math:`\lfloor \left(1 - c\right)m\rfloor` samples
            are selected.
        :type compression: float, optional
        :param optargs: Arguments for 'least_squares' optimizer.
            If set to None, `OPT_DEF_ARGS` are used as default parameters.
            Defaults to None.
        :type optargs: Dict[str, Any], optional
        """

        # super constructor not called
        # as most of the attributes are
        # not used.

        if optargs is None:
            optargs = OPT_DEF_ARGS

        self._Atilde = VarProOperator(
            svd_rank, exact, sorted_eigs, compression, optargs
        )
        self._optres: OptimizeResult = None
        self._snapshots_holder: Snapshots = None
        self._indices: np.ndarray = None
        self._modes_activation_bitmask_proxy = None

    def fit(self, X: np.ndarray, time: np.ndarray) -> object:
        r"""
        Fit the eigenvalues, modes and eigenfunctions/amplitudes
        to measurements.

        :param X: Measurements
            :math:`\boldsymbol{X} \in \mathbb{C}^{n \times m}`.
        :type X: np.ndarray
        :param time: 1d array of timestamps where measurements were taken.
        :type time: np.ndarray
        :return: VarProDMD instance.
        :rtype: object
        """

        self._snapshots_holder = Snapshots(X)
        (self._b, self._optres, self._indices) = self._Atilde.compute_operator(
            self._snapshots_holder.snapshots.astype(np.complex128), time
        )
        self._original_time = time
        self._dmd_time = time[self._indices]

        return self

    def forecast(self, time: np.ndarray) -> np.ndarray:
        r"""
        Forecast measurements at given timestamps `time`.

        :param time: Desired times for forcasting as 1d array.
        :type time: np.ndarray
        :raises ValueError: If method `fit(X, time)` was not called.
        :return: Predicted measurements :math:`\hat{\boldsymbol{X}}`.
        :rtype: np.ndarray
        """

        if not self.fitted:
            raise ValueError("Nothing fitted yet. Call fit-method first!")

        return varprodmd_predict(
            self._Atilde.modes, self._Atilde.eigenvalues, self._b, time
        )

    @property
    def ssr(self) -> float:
        """
        Compute the square root of sum squared residual (SSR) taken from
        https://link.springer.com/article/10.1007/s10589-012-9492-9.
        The SSR gives insight w.r.t. signal qualities.
        A low SSR is desired. If SSR is high the model may be inaccurate.

        :raises ValueError: ValueError is raised if method
            `fit(X, time)` was not called.
        :return: SSR.
        :rtype: float
        """

        if not self.fitted:
            raise ValueError("Nothing fitted yet!")

        rho_flat_real = self._optres.fun
        rho_flat_imag = np.zeros(
            (rho_flat_real.size // 2,), dtype=np.complex128
        )
        rho_flat_imag.real = rho_flat_real[: rho_flat_real.size // 2]
        rho_flat_imag.imag = rho_flat_real[rho_flat_real.size // 2 :]

        sigma = np.linalg.norm(rho_flat_imag)
        denom = max(
            self._original_time.size
            - self._optres.jac.shape[0] // 2
            - self._optres.jac.shape[1] // 2,
            1,
        )
        ssr = sigma / np.sqrt(float(denom))

        return ssr

    @property
    def selected_samples(self) -> np.ndarray:
        r"""
        Return indices for creating the library.

        :raises ValueError: ValueError is raised if method
            `fit(X, time)` was not called.
        :return: Indices of the selected samples.
            If no compression was performed :math:`\left(c = 0\right)`,
            all indices are returned, else indices of the
            library selection scheme using QR-Decomposition
            with column pivoting.
        :rtype: np.ndarray
        """

        if not self.fitted:
            raise ValueError("Nothing fitted yet. Call fit-method first!")

        return self._indices

    @property
    def opt_stats(self) -> OptimizeResult:
        """
        Return optimization statistics of the Variable Projection
        optimization.

        :raises ValueError: ValueError is raised if method `fit(X, time)`
            was not called.
        :return: Optimization results including optimal weights
            (continuous eigenvalues) and number of iterations.
        :rtype: OptimizeResult
        """

        if not self.fitted:
            raise ValueError("Nothing fitted yet!")

        return self._optres

    @property
    def dynamics(self):
        """
        Get the time evolution of each mode.

        :return: matrix that contains all the time evolution, stored by row.
        :rtype: numpy.ndarray
        """

        t_omega = np.exp(np.outer(self.eigs, self._original_time))
        return self.amplitudes[:, None] * t_omega

    @property
    def frequency(self):
        """
        Get the amplitude spectrum.

        :return: the array that contains the frequencies of the eigenvalues.
        :rtype: numpy.ndarray
        """

        return self.eigs.imag / (2 * np.pi)

    @property
    def growth_rate(self):
        """
        Get the growth rate values relative to the modes.

        :return: the Floquet values
        :rtype: numpy.ndarray
        """

        return self.eigs.real



"""
Derived module from hankeldmd.py for higher order dmd.

Reference:
- S. L Clainche, J. M. Vega, Higher Order Dynamic Mode Decomposition.
Journal on Applied Dynamical Systems, 16(2), 882-925, 2017.
"""

import warnings

import numpy as np

from .hankeldmd import HankelDMD
from .preprocessing.svd_projection import svd_projection_preprocessing
from .snapshots import Snapshots


class HODMD(HankelDMD):
    """
    Higher Order Dynamic Mode Decomposition

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means no truncation.
    :param bool exact: flag to compute either exact DMD or projected DMD.
        Default is False.
    :param opt: argument to control the computation of DMD modes amplitudes.
        See :class:`DMDBase`. Default is False.
    :type opt: bool or int
    :param rescale_mode: Scale Atilde as shown in
            10.1016/j.jneumeth.2015.10.010 (section 2.4) before computing its
            eigendecomposition. None means no rescaling, 'auto' means automatic
            rescaling using singular values, otherwise the scaling factors.
    :type rescale_mode: {'auto'} or None or numpy.ndarray
    :param bool forward_backward: If True, the low-rank operator is computed
        like in fbDMD (reference: https://arxiv.org/abs/1507.02264). Default is
        False.
    :param int d: the new order for spatial dimension of the input snapshots.
        Default is 1.
    :param sorted_eigs: Sort eigenvalues (and modes/dynamics accordingly) by
        magnitude if `sorted_eigs='abs'`, by real part (and then by imaginary
        part to break ties) if `sorted_eigs='real'`. Default: False.
    :type sorted_eigs: {'real', 'abs'} or False
    :param reconstruction_method: Method used to reconstruct the snapshots of
        the dynamical system from the multiple versions available due to how
        HODMD is conceived. If `'first'` (default) the first version
        available is selected (i.e. the nearest to the 0-th row in the
        augmented matrix). If `'mean'` we compute the element-wise mean. If
        `reconstruction_method` is an array of float values we compute the
        weighted average (for each snapshots) using the given values as weights
        (the number of weights must be equal to `d`).
    :param svd_rank_extra: the rank for the initial reduction of the input
        data, performed before the rearrangement of the input data to the
        (pseudo) Hankel matrix format; If 0, the method computes the optimal
        rank and uses it for truncation; if positive interger, the method uses
        the argument for the truncation; if float between 0 and 1, the rank is
        the number of the biggest singular values that are needed to reach the
        'energy' specified by `svd_rank`; if -1, the method does not compute
        truncation.
    :type svd_rank: int or float
    """

    def __init__(
        self,
        svd_rank=0,
        tlsq_rank=0,
        exact=False,
        opt=False,
        rescale_mode=None,
        forward_backward=False,
        d=1,
        sorted_eigs=False,
        reconstruction_method="first",
        svd_rank_extra=0,
    ):
        super().__init__(
            svd_rank=svd_rank,
            tlsq_rank=tlsq_rank,
            exact=exact,
            opt=opt,
            rescale_mode=rescale_mode,
            forward_backward=forward_backward,
            d=d,
            sorted_eigs=sorted_eigs,
            reconstruction_method=reconstruction_method,
        )
        self._sub_dmd = svd_projection_preprocessing(
            self._sub_dmd, svd_rank_extra
        )

    def fit(self, X):
        """
        Compute the Dynamic Modes Decomposition to the input data.

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        """
        snapshots_holder = Snapshots(X)

        super().fit(snapshots_holder.snapshots)
        self._snapshots_holder = snapshots_holder

        return self



import os

import matplotlib.pyplot as plt
import numpy as np
from pytest import raises

from pydmd import DMD, DMDBase, MrDMD
from pydmd.plotter import (
    _enforce_ratio,
    _plot_limits,
    plot_eigs,
    plot_eigs_mrdmd,
    plot_modes_2D,
    plot_snapshots_2D,
    plot_summary,
)

from .test_mrdmd import sample_data as mrdmd_sample_data

sample_data = np.load("tests/test_datasets/input_sample.npy")


def test_enforce_ratio_y():
    supx, infx, supy, infy = _enforce_ratio(10, 20, 10, 0, 0)

    dx = supx - infx
    dy = supy - infy
    np.testing.assert_almost_equal(max(dx, dy) / min(dx, dy), 10, decimal=6)


def test_enforce_ratio_x():
    supx, infx, supy, infy = _enforce_ratio(10, 0, 0, 20, 10)

    dx = supx - infx
    dy = supy - infy
    np.testing.assert_almost_equal(max(dx, dy) / min(dx, dy), 10, decimal=6)


def test_plot_limits_narrow():
    dmd = DMDBase()
    dmd.operator._eigenvalues = np.array([complex(1, 2), complex(-1, -2)])
    dmd.operator._modes = np.array(np.ones((10, 2)))

    tp = _plot_limits(dmd, True)

    assert len(tp) == 4

    supx, infx, supy, infy = tp
    assert supx == 1.05
    assert infx == -1.05
    assert supy == 2.05
    assert infy == -2.05


def test_plot_limits():
    dmd = DMDBase()
    dmd.operator._eigenvalues = np.array([complex(-2, 2), complex(3, -3)])
    dmd.operator._modes = np.array(np.ones((10, 2)))

    limit = _plot_limits(dmd, False)
    assert limit == 5


def test_plot_eigs():
    dmd = DMDBase()
    with raises(ValueError):
        plot_eigs(dmd, show_axes=True, show_unit_circle=True)


def test_plot_eigs_narrowview_empty():
    dmd = DMDBase()
    # max/min throws an error if the array is empty (max used on empty
    # array)
    dmd.operator._eigenvalues = np.array([], dtype=complex)
    with raises(ValueError):
        plot_eigs(dmd, show_axes=False, narrow_view=True, dpi=200)


def test_plot_modes_2D():
    dmd = DMDBase()
    with raises(ValueError):
        plot_modes_2D(dmd)


def test_plot_snaps_2D():
    dmd = DMDBase()
    with raises(ValueError):
        plot_snapshots_2D(dmd)


def test_plot_eigs_1():
    dmd = DMD()
    dmd.fit(X=sample_data)
    plot_eigs(dmd, show_axes=True, show_unit_circle=True)
    plt.close()


def test_plot_eigs_2():
    dmd = DMD()
    dmd.fit(X=sample_data)
    plot_eigs(dmd, show_axes=False, show_unit_circle=False)
    plt.close()


def test_plot_eigs_3():
    dmd = DMD()
    dmd.fit(X=sample_data)
    plot_eigs(dmd, show_axes=False, show_unit_circle=True, filename="eigs.png")
    os.remove("eigs.png")


def test_plot_modes_1():
    dmd = DMD()
    dmd.fit(X=sample_data)
    with raises(ValueError):
        plot_modes_2D(dmd)


def test_plot_modes_2():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    plot_modes_2D(dmd, index_mode=(1, 2, 5), x=np.arange(20), y=np.arange(20))
    plt.close()


def test_plot_modes_3():
    dmd = DMD()
    snapshots = [snap.reshape(20, 20) for snap in sample_data.T]
    dmd.fit(X=snapshots)
    plot_modes_2D(dmd, snapshots_shape=(20, 20))
    plt.close()


def test_plot_modes_4():
    dmd = DMD()
    snapshots = [snap.reshape(20, 20) for snap in sample_data.T]
    dmd.fit(X=snapshots)
    plot_modes_2D(dmd, snapshots_shape=(20, 20), index_mode=1)
    plt.close()


def test_plot_modes_5():
    dmd = DMD()
    snapshots = [snap.reshape(20, 20) for snap in sample_data.T]
    dmd.fit(X=snapshots)
    plot_modes_2D(
        dmd, snapshots_shape=(20, 20), index_mode=1, filename="tmp.png"
    )
    os.remove("tmp.1.png")


def test_plot_snapshots_1():
    dmd = DMD()
    dmd.fit(X=sample_data)
    with raises(ValueError):
        plot_snapshots_2D(dmd)


def test_plot_snapshots_2():
    dmd = DMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    plot_snapshots_2D(dmd, (1, 2, 5), x=np.arange(20), y=np.arange(20))
    plt.close()


def test_plot_snapshots_3():
    dmd = DMD()
    snapshots = [snap.reshape(20, 20) for snap in sample_data.T]
    dmd.fit(X=snapshots)
    plot_snapshots_2D(dmd, snapshots_shape=(20, 20))
    plt.close()


def test_plot_snapshots_4():
    dmd = DMD()
    snapshots = [snap.reshape(20, 20) for snap in sample_data.T]
    dmd.fit(X=snapshots)
    plot_snapshots_2D(dmd, snapshots_shape=(20, 20), index_snap=2)
    plt.close()


def test_plot_snapshots_5():
    dmd = DMD()
    snapshots = [snap.reshape(20, 20) for snap in sample_data.T]
    dmd.fit(X=snapshots)
    plot_snapshots_2D(
        dmd, snapshots_shape=(20, 20), index_snap=2, filename="tmp.png"
    )
    os.remove("tmp.2.png")


def test_tdmd_plot():
    dmd = DMD(tlsq_rank=3)
    dmd.fit(X=sample_data)
    plot_eigs(dmd, show_axes=False, show_unit_circle=False)
    plt.close()


def test_mrdmd_wrong_plot_eig1():
    rank = 2
    dmd = DMD(svd_rank=rank)
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=mrdmd_sample_data)
    with raises(ValueError):
        plot_eigs_mrdmd(
            mrdmd,
            show_axes=True,
            show_unit_circle=True,
            figsize=(8, 8),
            level=7,
        )


def test_mrdmd_plot_eig1():
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=mrdmd_sample_data)
    plot_eigs_mrdmd(
        mrdmd, show_axes=True, show_unit_circle=True, figsize=(8, 8)
    )
    plt.close()


def test_mrdmd_plot_eig2():
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=mrdmd_sample_data)
    plot_eigs_mrdmd(
        mrdmd, show_axes=True, show_unit_circle=False, title="Title"
    )
    plt.close()


def test_mrdmd_plot_eig3():
    dmd = DMD()
    mrdmd = MrDMD(dmd, max_level=6, max_cycles=2)
    mrdmd.fit(X=mrdmd_sample_data)
    plot_eigs_mrdmd(
        mrdmd, show_axes=False, show_unit_circle=False, level=1, node=0
    )
    plt.close()


def test_plot_summary_1():
    # Test 1: Throw error if more than 3 modes are requested.
    dmd = DMD()
    dmd.fit(X=sample_data)
    with raises(ValueError):
        plot_summary(dmd, index_modes=[0, 1, 2, 3])


def test_plot_summary_2():
    # Test 2: Everything is fine with all default settings.
    dmd = DMD()
    dmd.fit(X=sample_data)
    plot_summary(dmd)
    plt.close()


def test_plot_summary_3():
    # Test 3: Everything is fine when plotting in 2D.
    dmd = DMD()
    dmd.fit(X=sample_data)
    plot_summary(dmd, snapshots_shape=(20, 20))
    plt.close()


def test_plot_summary_4():
    # Test 4: Everything is fine when plotting fewer than 3 modes.
    dmd = DMD(svd_rank=2)
    dmd.fit(X=sample_data)
    plot_summary(dmd, snapshots_shape=(20, 20))
    plt.close()


def test_plot_summary_5():
    # Test 5: Everything is fine when saving the plot.
    dmd = DMD()
    dmd.fit(X=sample_data)
    plot_summary(dmd, snapshots_shape=(20, 20), filename="tmp.png")
    os.remove("tmp.png")



import os
from builtins import range

import matplotlib.pyplot as plt
import numpy as np
from pytest import raises

from pydmd.dmdoperator import DMDOperator
from pydmd.utils import compute_tlsq

# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")


def test_constructor():
    operator = DMDOperator(
        svd_rank=2,
        exact=True,
        forward_backward=False,
        rescale_mode="auto",
        sorted_eigs=False,
        tikhonov_regularization=None,
    )

    assert operator._svd_rank == 2
    assert operator._exact == True
    assert operator._forward_backward == False
    assert operator._rescale_mode == "auto"


def test_noncompute_error():
    operator = DMDOperator(
        svd_rank=2,
        exact=True,
        forward_backward=False,
        rescale_mode="auto",
        sorted_eigs=False,
        tikhonov_regularization=None,
    )

    with raises(ValueError):
        operator.shape

    with raises(ValueError):
        operator.Lambda

    with raises(ValueError):
        operator.modes

    with raises(ValueError):
        operator.eigenvalues

    with raises(ValueError):
        operator.eigenvectors

    with raises(ValueError):
        operator.as_numpy_array


def test_compute_operator():
    operator = DMDOperator(
        svd_rank=0,
        exact=True,
        forward_backward=False,
        rescale_mode="auto",
        sorted_eigs=False,
        tikhonov_regularization=None,
    )
    operator.compute_operator(np.ones((3, 3)), np.ones((3, 3)))

    assert operator.as_numpy_array is not None
    assert operator.eigenvalues is not None
    assert operator.eigenvectors is not None
    assert operator.modes is not None
    assert operator.Lambda is not None


# test that a value of 'auto' in rescale_mode is replaced by the singular
# values of X
def test_rescalemode_auto_singular_values():
    operator = DMDOperator(
        svd_rank=0,
        exact=True,
        forward_backward=False,
        rescale_mode="auto",
        sorted_eigs=False,
        tikhonov_regularization=None,
    )
    operator.compute_operator(np.ones((3, 3)), np.ones((3, 3)))
    np.testing.assert_almost_equal(
        operator._rescale_mode, np.array([3.0]), decimal=1
    )


def test_call():
    operator = DMDOperator(
        svd_rank=2,
        exact=True,
        forward_backward=False,
        rescale_mode=None,
        sorted_eigs=False,
        tikhonov_regularization=None,
    )

    X = sample_data[:, :-1]
    Y = sample_data[:, 1:]
    X, Y = compute_tlsq(X, Y, 0)

    operator.compute_operator(X, Y)

    expected = np.array([-0.47643628 + 0.87835227j, -0.47270971 + 0.88160808j])

    np.testing.assert_almost_equal(operator(np.ones(2)), expected, decimal=6)


def test_compute_eigenquantities_wrong_rescalemode():
    operator = DMDOperator(
        svd_rank=0,
        exact=True,
        forward_backward=False,
        rescale_mode=4,
        sorted_eigs=False,
        tikhonov_regularization=None,
    )
    with raises(ValueError):
        operator.compute_operator(np.ones((3, 3)), np.ones((3, 3)))

    operator = DMDOperator(
        svd_rank=0,
        exact=True,
        forward_backward=False,
        rescale_mode=np.ones((4,)),
        sorted_eigs=False,
        tikhonov_regularization=None,
    )
    with raises(ValueError):
        operator.compute_operator(np.ones((3, 3)), np.ones((3, 3)))



from builtins import range

import numpy as np
from pytest import raises

from pydmd.optdmd import OptDMD

# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")


def create_noisy_data():
    mu = 0.0
    sigma = 0.0  # noise standard deviation
    m = 100  # number of snapshot
    noise = np.random.normal(mu, sigma, m)  # gaussian noise
    A = np.array([[1.0, 1.0], [-1.0, 2.0]])
    A /= np.sqrt(3)
    n = 2
    X = np.zeros((n, m))
    X[:, 0] = np.array([0.5, 1.0])
    # evolve the system and perturb the data with noise
    for k in range(1, m):
        X[:, k] = A.dot(X[:, k - 1])
        X[:, k - 1] += noise[k - 1]
    return X


noisy_data = create_noisy_data()


def test_shape_1():
    optdmd = OptDMD(svd_rank=-1)
    optdmd.fit(X=sample_data)
    assert optdmd.modes.shape[1] == sample_data.shape[1] - 1


def test_shape_2():
    optdmd = OptDMD(svd_rank=-1)
    optdmd.fit(X=sample_data[:, :-1], Y=sample_data[:, 1:])
    assert optdmd.modes.shape[1] == sample_data.shape[1] - 1


def test_truncation_shape():
    optdmd = OptDMD(svd_rank=3)
    optdmd.fit(X=sample_data)
    assert optdmd.modes.shape[1] == 3


def test_rank():
    optdmd = OptDMD(svd_rank=0.9)
    optdmd.fit(X=sample_data)
    assert len(optdmd.eigs) == 2


def test_Atilde_shape():
    optdmd = OptDMD(svd_rank=3)
    optdmd.fit(X=sample_data)
    assert optdmd.operator.as_numpy_array.shape == (
        optdmd.operator._svd_rank,
        optdmd.operator._svd_rank,
    )


def test_Atilde_values():
    optdmd = OptDMD(svd_rank=2)
    optdmd.fit(X=sample_data)
    exact_atilde = np.array(
        [
            [-0.70558526 + 0.67815084j, 0.22914898 + 0.20020143j],
            [0.10459069 + 0.09137814j, -0.57730040 + 0.79022994j],
        ]
    )
    np.testing.assert_allclose(
        np.linalg.eigvals(exact_atilde),
        np.linalg.eigvals(optdmd.operator.as_numpy_array),
    )


def test_eigs_1():
    optdmd = OptDMD(svd_rank=-1)
    optdmd.fit(X=sample_data)
    assert len(optdmd.eigs) == 14


def test_eigs_2():
    optdmd = OptDMD(svd_rank=5)
    optdmd.fit(X=sample_data)
    assert len(optdmd.eigs) == 5


def test_eigs_3():
    optdmd = OptDMD(svd_rank=2)
    optdmd.fit(X=sample_data)
    expected_eigs = np.array(
        [-8.09016994e-01 + 5.87785252e-01j, -4.73868662e-01 + 8.80595532e-01j]
    )
    np.testing.assert_almost_equal(optdmd.eigs, expected_eigs, decimal=6)

    # def test_dynamics_1():
    #     dmd = DMD(svd_rank=5)
    #     dmd.fit(X=sample_data)
    #     assert dmd.dynamics.shape == (5, sample_data.shape[1])
    #
    # def test_dynamics_2():
    #     dmd = DMD(svd_rank=1)
    #     dmd.fit(X=sample_data)
    #     expected_dynamics = np.array([[
    #         -2.20639502 - 9.10168802e-16j, 1.55679980 - 1.49626864e+00j,
    #         -0.08375915 + 2.11149018e+00j, -1.37280962 - 1.54663768e+00j,
    #         2.01748787 + 1.60312745e-01j, -1.53222592 + 1.25504678e+00j,
    #         0.23000498 - 1.92462280e+00j, 1.14289644 + 1.51396355e+00j,
    #         -1.83310653 - 2.93174173e-01j, 1.49222925 - 1.03626336e+00j,
    #         -0.35015209 + 1.74312867e+00j, -0.93504202 - 1.46738182e+00j,
    #         1.65485808 + 4.01263449e-01j, -1.43976061 + 8.39117825e-01j,
    #         0.44682540 - 1.56844403e+00j
    #     ]])
    #     np.testing.assert_allclose(dmd.dynamics, expected_dynamics)
    #
    # def test_dynamics_opt_1():
    #     dmd = DMD(svd_rank=5, opt=True)
    #     dmd.fit(X=sample_data)
    #     assert dmd.dynamics.shape == (5, sample_data.shape[1])
    #
    # def test_dynamics_opt_2():
    #     dmd = DMD(svd_rank=1, opt=True)
    #     dmd.fit(X=sample_data)
    #     expected_dynamics = np.array([[
    #         -4.56004133 - 6.48054238j, 7.61228319 + 1.4801793j,
    #         -6.37489962 + 4.11788355j, 1.70548899 - 7.22866146j,
    #         3.69875496 + 6.25701574j, -6.85298745 - 1.90654427j,
    #         6.12829151 - 3.30212967j, -2.08469012 + 6.48584004j,
    #         -2.92745126 - 5.99004747j, 6.12772217 + 2.24123565j,
    #         -5.84352626 + 2.57413711j, 2.37745273 - 5.77906544j,
    #         2.24158249 + 5.68989493j, -5.44023459 - 2.49457492j,
    #         5.53024740 - 1.92916437j
    #     ]])
    #     np.testing.assert_allclose(dmd.dynamics, expected_dynamics)

    # def test_reconstructed_data():
    #     dmd = DMD()
    #     dmd.fit(X=sample_data)
    #     dmd_data = dmd.reconstructed_data
    #     np.testing.assert_allclose(dmd_data, sample_data)

    # def test_original_time():
    #     dmd = DMD(svd_rank=2)
    #     dmd.fit(X=sample_data)
    #     expected_dict = {'dt': 1, 't0': 0, 'tend': 14}
    #     np.testing.assert_equal(dmd.original_time, expected_dict)
    #
    # def test_original_timesteps():
    #     dmd = DMD()
    #     dmd.fit(X=sample_data)
    #     np.testing.assert_allclose(dmd.original_timesteps,
    #                                np.arange(sample_data.shape[1]))
    #
    # def test_dmd_time_1():
    #     dmd = DMD(svd_rank=2)
    #     dmd.fit(X=sample_data)
    #     expected_dict = {'dt': 1, 't0': 0, 'tend': 14}
    #     np.testing.assert_equal(dmd.dmd_time, expected_dict)
    #
    # def test_dmd_time_2():
    #     dmd = DMD()
    #     dmd.fit(X=sample_data)
    #     dmd.dmd_time['t0'] = 10
    #     dmd.dmd_time['tend'] = 14
    #     expected_data = sample_data[:, -5:]
    #     np.testing.assert_allclose(dmd.reconstructed_data, expected_data)
    #
    # def test_dmd_time_3():
    #     dmd = DMD()
    #     dmd.fit(X=sample_data)
    #     dmd.dmd_time['t0'] = 8
    #     dmd.dmd_time['tend'] = 11
    #     expected_data = sample_data[:, 8:12]
    #     np.testing.assert_allclose(dmd.reconstructed_data, expected_data)
    #
    # def test_dmd_time_4():
    #     dmd = DMD(svd_rank=3)
    #     dmd.fit(X=sample_data)
    #     dmd.dmd_time['t0'] = 20
    #     dmd.dmd_time['tend'] = 20
    #     expected_data = np.array([[-7.29383297e+00 - 4.90248179e-14j],
    #                               [-5.69109796e+00 - 2.74068833e+00j],
    #                               [3.38410649e-83 + 3.75677740e-83j]])
    #     np.testing.assert_almost_equal(dmd.dynamics, expected_data, decimal=6)
    #


def test_bitmask_not_implemented():
    with raises(RuntimeError):
        optdmd = OptDMD(svd_rank=2)
        optdmd.fit(X=sample_data)
        optdmd.modes_activation_bitmask
    with raises(RuntimeError):
        optdmd = OptDMD(svd_rank=2)
        optdmd.fit(X=sample_data)
        optdmd.modes_activation_bitmask = None


def test_getitem_not_implemented():
    with raises(RuntimeError):
        optdmd = OptDMD(svd_rank=2)
        optdmd.fit(X=sample_data)
        optdmd[1:3]



import numpy as np
import pytest

from pydmd.snapshots import Snapshots


def test_1d():
    with pytest.raises(ValueError):
        Snapshots(np.ones(10))


def test_2d():
    X = np.random.rand(10, 5)
    snapshots = Snapshots(X)
    assert np.all(snapshots.snapshots == X)
    assert snapshots.snapshots_shape == (10,)


def test_3d():
    X = np.random.rand(10, 3, 5)
    snapshots = Snapshots(X)
    assert np.all(snapshots.snapshots == X.reshape((-1, 5)))
    assert snapshots.snapshots_shape == (10, 3)


def test_4d():
    X = np.random.rand(10, 3, 2, 5)
    snapshots = Snapshots(X)
    assert np.all(snapshots.snapshots == X.reshape((-1, 5)))
    assert snapshots.snapshots_shape == (10, 3, 2)


def test_list_1d():
    X = [1, 2, 3, 4]
    with pytest.raises(ValueError):
        Snapshots(X)


def test_list_2d():
    X = [[1, 2], [2, 3], [3, 4], [4, 5]]
    snapshots = Snapshots(X)
    assert snapshots.snapshots_shape == (2,)
    assert np.all(snapshots.snapshots == [[1, 2, 3, 4], [2, 3, 4, 5]])


def test_list_3d():
    X = [
        [[1, 2, 0], [2, 3, 0]],
        [[3, 4, 0], [4, 5, 0]],
        [[5, 6, 0], [6, 7, 0]],
        [[7, 8, 0], [8, 9, 0]],
    ]
    snapshots = Snapshots(X)
    assert snapshots.snapshots_shape == (2, 3)
    assert np.all(
        snapshots.snapshots
        == [
            [1, 3, 5, 7],
            [2, 4, 6, 8],
            [0, 0, 0, 0],
            [2, 4, 6, 8],
            [3, 5, 7, 9],
            [0, 0, 0, 0],
        ]
    )



"""
Test module for VarProDMD
"""

import numpy as np
import pytest

from pydmd import VarProDMD
from pydmd.varprodmd import (
    OPT_DEF_ARGS,
    _compute_dmd_jac,
    _compute_dmd_rho,
    _OptimizeHelper,
    compute_varprodmd_any,
    varprodmd_predict,
    select_best_samples_fast,
)


def signal(x_loc: np.ndarray, time: np.ndarray) -> np.ndarray:
    """
    construct complex spatio temporal signal for testing.
    :param x_loc: 1d x-coordinate.
    :type x_loc: np.ndarray
    :param time: 1d time array.
    :type time: np.ndarray
    :return: Spatiotemporal signal.
    :rtype: np.ndarray
    """
    f_1 = 1.0 / np.cosh(x_loc + 3) * np.exp(1j * 2.3 * time)
    f_2 = 2.0 / np.cosh(x_loc) * np.tanh(x_loc) * np.exp(1j * 2.8 * time)
    return f_1 + f_2


def test_varprodmd_rho():
    """
    Unit test for residual vector :math: `\boldsymbol{\rho}`.
    """
    data = np.eye(2, 2).astype(np.complex128)
    time = np.array([0.0, 1.0], np.float64)
    alphas = np.array([1.0 + 0j, 1.0 + 0j], np.complex128)
    alphas_in = np.array([1.0, 1.0, 0.0, 0.0], np.float64)
    phi = np.exp(np.outer(time, alphas))
    U_svd, s_svd, V_svd_t = np.linalg.svd(
        phi, hermitian=False, full_matrices=False
    )
    idx = np.where(s_svd.real != 0.0)[0]
    s_inv = np.zeros_like(s_svd)
    s_inv[idx] = np.reciprocal(s_svd[idx])

    res = data - np.linalg.multi_dot([U_svd, U_svd.conj().T, data])
    res_flat = np.ravel(res)
    res_flat_reals = np.zeros((2 * res_flat.shape[-1]))
    res_flat_reals[: res_flat_reals.shape[-1] // 2] = res_flat.real
    res_flat_reals[res_flat_reals.shape[-1] // 2 :] = res_flat.imag
    opthelper = _OptimizeHelper(2, *data.shape)
    rho_flat_out = _compute_dmd_rho(alphas_in, time, data, opthelper)

    assert np.array_equal(rho_flat_out, res_flat_reals)
    assert np.array_equal(U_svd, opthelper.u_svd)
    assert np.array_equal(s_inv, opthelper.s_inv)
    assert np.array_equal(V_svd_t.conj().T, opthelper.v_svd)
    assert np.array_equal(phi, opthelper.phi)


def test_varprodmd_jac():  # pylint: disable=too-many-locals,too-many-statements
    """
    Test Jacobian computation (real vs. complex).
    """
    data = np.eye(2, 2).astype(np.complex128)
    time = np.array([0.0, 1.0])
    alphas = np.array([-1.0 + 0.0j, -2.0 + 0.0j], np.complex128)
    alphas_in = np.array([-1.0, -2.0, 0.0, 0.0], np.float64)
    phi = np.exp(np.outer(time, alphas))
    d_phi_1 = np.zeros((2, 2), dtype=np.complex128)
    d_phi_2 = np.zeros((2, 2), dtype=np.complex128)
    d_phi_1[:, 0] = time * phi[:, 0]
    d_phi_2[:, 1] = time * phi[:, 1]

    U_svd, s_svd, __v = np.linalg.svd(phi, hermitian=False, full_matrices=False)
    idx = np.where(s_svd.real != 0.0)[0]
    s_inv = np.zeros_like(s_svd)
    s_inv[idx] = np.reciprocal(s_svd[idx])
    phi_inv = (__v.conj().T * s_inv.reshape((1, -1))) @ U_svd.conj().T

    opthelper = _OptimizeHelper(2, *data.shape)
    opthelper.u_svd = U_svd
    opthelper.v_svd = __v.conj().T
    opthelper.s_inv = s_inv
    opthelper.phi = phi
    opthelper.phi_inv = phi_inv
    opthelper.b_matrix = phi_inv @ data
    opthelper.rho = data - phi @ opthelper.b_matrix
    rho_flat = np.ravel(opthelper.rho)
    rho_real = np.zeros((2 * rho_flat.shape[0]))
    rho_real[: rho_flat.shape[0]] = rho_flat.real
    rho_real[rho_flat.shape[0] :] = rho_flat.imag
    A_1 = d_phi_1 @ opthelper.b_matrix - np.linalg.multi_dot(
        [U_svd, U_svd.conj().T, d_phi_1, opthelper.b_matrix]
    )

    A_2 = d_phi_2 @ opthelper.b_matrix - np.linalg.multi_dot(
        [U_svd, U_svd.conj().T, d_phi_2, opthelper.b_matrix]
    )

    G_1 = np.linalg.multi_dot(
        [phi_inv.conj().T, d_phi_1.conj().T, opthelper.rho]
    )
    G_2 = np.linalg.multi_dot(
        [phi_inv.conj().T, d_phi_2.conj().T, opthelper.rho]
    )
    J_1 = -A_1 - G_1
    J_2 = -A_2 - G_2
    J_1_flat = np.ravel(J_1)
    J_2_flat = np.ravel(J_2)
    JAC_IMAG = np.zeros((J_1_flat.shape[0], 2), dtype=np.complex128)
    JAC_IMAG[:, 0] = J_1_flat
    JAC_IMAG[:, 1] = J_2_flat
    JAC_REAL = np.zeros((2 * J_1_flat.shape[-1], 4), dtype=np.float64)
    JAC_REAL[: J_1_flat.shape[-1], 0] = J_1_flat.real
    JAC_REAL[J_1_flat.shape[-1] :, 0] = J_1_flat.imag
    JAC_REAL[: J_2_flat.shape[-1], 1] = J_2_flat.real
    JAC_REAL[J_2_flat.shape[-1] :, 1] = J_2_flat.imag
    JAC_REAL[: J_1_flat.shape[-1], 2] = -J_1_flat.imag
    JAC_REAL[J_1_flat.shape[-1] :, 2] = J_1_flat.real
    JAC_REAL[: J_2_flat.shape[-1], 3] = -J_2_flat.imag
    JAC_REAL[J_2_flat.shape[-1] :, 3] = J_2_flat.real
    JAC_OUT_REAL = _compute_dmd_jac(alphas_in, time, data, opthelper)

    GRAD_REAL = JAC_REAL.T @ rho_real
    GRAD_OUT_REAL = JAC_OUT_REAL.T @ rho_real
    GRAD_IMAG = JAC_IMAG.conj().T @ rho_flat

    assert np.linalg.norm(JAC_REAL - JAC_OUT_REAL) < 1e-12
    assert np.linalg.norm(GRAD_REAL - GRAD_OUT_REAL) < 1e-12

    imag2real = np.zeros_like(GRAD_REAL)
    imag2real[: imag2real.shape[-1] // 2] = GRAD_IMAG.real
    imag2real[imag2real.shape[-1] // 2 :] = GRAD_IMAG.imag

    rec_grad = np.zeros_like(GRAD_IMAG)
    rec_grad.real = GRAD_REAL[: GRAD_REAL.shape[-1] // 2]
    rec_grad.imag = GRAD_REAL[GRAD_REAL.shape[-1] // 2 :]

    # funny numerical errors leads to
    # np.array_equal(GRAD_IMAG, __rec_grad) to fail
    assert np.linalg.norm(GRAD_IMAG - rec_grad) < 1e-9

    rec_grad = np.zeros_like(GRAD_IMAG)
    rec_grad.real = GRAD_OUT_REAL[: GRAD_OUT_REAL.shape[-1] // 2]
    rec_grad.imag = GRAD_OUT_REAL[GRAD_OUT_REAL.shape[-1] // 2 :]

    # funny numerical errors leads to
    # np.array_equal(GRAD_IMAG, __rec_grad) to fail
    assert np.linalg.norm(GRAD_IMAG - rec_grad) < 1e-9


def test_varprodmd_any():
    """
    Test Variable Projection function for DMD (at any timestep).
    """
    time = np.linspace(0, 4 * np.pi, 100)
    x_loc = np.linspace(-10, 10, 1024)

    z = signal(*np.meshgrid(x_loc, time)).T

    idx = select_best_samples_fast(z, 0.6)

    z_sub = z[:, idx]
    t_sub = time[idx]

    with pytest.raises(ValueError):
        compute_varprodmd_any(
            z_sub[:, 0],
            t_sub,
            OPT_DEF_ARGS,
            rank=0.0,
        )

    with pytest.raises(ValueError):
        compute_varprodmd_any(
            z_sub, t_sub.reshape((-1, 1)), OPT_DEF_ARGS, rank=0.0
        )

    phi, lambdas, eigenf, _, _ = compute_varprodmd_any(
        z_sub, t_sub, OPT_DEF_ARGS, rank=0.0
    )
    pred = varprodmd_predict(phi, lambdas, eigenf, time)
    diff = np.abs(pred - z)
    mae = np.sum(np.sum(diff, axis=0), axis=-1) / z.shape[0] / z.shape[-1]

    assert mae < 1.0

    phi, lambdas, eigenf, _, _ = compute_varprodmd_any(
        z_sub, t_sub, OPT_DEF_ARGS, rank=0.0, use_proj=False
    )
    pred = varprodmd_predict(phi, lambdas, eigenf, time)
    diff = np.abs(pred - z)
    mae = np.sum(np.sum(diff, axis=0), axis=-1) / z.shape[0] / z.shape[-1]

    assert mae < 1.0


def test_varprodmd_class():
    """
    Test VarProDMD class.
    """
    time = np.linspace(0, 4 * np.pi, 100)
    x_loc = np.linspace(-10, 10, 1024)

    z = signal(*np.meshgrid(x_loc, time)).T
    dmd = VarProDMD(0, False, False, 0)

    with pytest.raises(ValueError):
        _ = dmd.forecast(time)

    with pytest.raises(ValueError):
        _ = dmd.ssr

    with pytest.raises(ValueError):
        _ = dmd.selected_samples

    with pytest.raises(ValueError):
        _ = dmd.opt_stats

    dmd.fit(z, time)
    assert dmd.fitted
    assert dmd.eigs.size > 0
    assert len(dmd.modes.shape) == 2
    assert dmd.amplitudes.size > 0
    assert len(dmd.dynamics.shape) == 2
    assert dmd.amplitudes.size == dmd.frequency.size
    assert dmd.growth_rate.size == dmd.amplitudes.size
    assert dmd.eigs.size == dmd.amplitudes.size

    pred = dmd.forecast(time)

    diff = np.abs(pred - z)
    mae = np.sum(np.sum(diff, axis=0), axis=-1) / z.shape[0] / z.shape[-1]

    assert mae < 1
    assert dmd.ssr < 1e-3

    dmd = VarProDMD(0, False, "unkown_sort", 0.8)

    with pytest.raises(ValueError):
        dmd.fit(z, time)

    sort_args = ["auto", "real", "imag", "abs", True, False]

    for arg in sort_args:
        dmd = VarProDMD(0, False, arg, 0.6)
        dmd.fit(z, time)
        pred = dmd.forecast(time)
        diff = np.abs(pred - z)
        mae = np.sum(np.sum(diff, axis=0), axis=-1) / z.shape[0] / z.shape[-1]
        assert dmd.selected_samples.size == int((1 - 0.6) * 100)
        assert mae < 1.0



"""
Derived module from dmdbase.py for classic dmd.
"""

import numpy as np
from scipy.linalg import pinv

from .dmdbase import DMDBase
from .snapshots import Snapshots
from .utils import compute_tlsq


class DMD(DMDBase):
    """
    Dynamic Mode Decomposition

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param int tlsq_rank: rank truncation computing Total Least Square. Default
        is 0, that means TLSQ is not applied.
    :param bool exact: flag to compute either exact DMD or projected DMD.
        Default is False.
    :param opt: argument to control the computation of DMD modes amplitudes.
        See :class:`DMDBase`. Default is False.
    :type opt: bool or int
    :param rescale_mode: Scale Atilde as shown in
            10.1016/j.jneumeth.2015.10.010 (section 2.4) before computing its
            eigendecomposition. None means no rescaling, 'auto' means automatic
            rescaling using singular values, otherwise the scaling factors.
    :type rescale_mode: {'auto'} or None or numpy.ndarray
    :param bool forward_backward: If True, the low-rank operator is computed
        like in fbDMD (reference: https://arxiv.org/abs/1507.02264). Default is
        False.
    :param sorted_eigs: Sort eigenvalues (and modes/dynamics accordingly) by
        magnitude if `sorted_eigs='abs'`, by real part (and then by imaginary
        part to break ties) if `sorted_eigs='real'`. Default: False.
    :type sorted_eigs: {'real', 'abs'} or False
    :param tikhonov_regularization: Tikhonov parameter for the regularization.
        If `None`, no regularization is applied, if `float`, it is used as the
        :math:`\lambda` tikhonov parameter.
    :type tikhonov_regularization: int or float
    """

    def fit(self, X, Y=None):
        """
        Compute the Dynamic Modes Decomposition to the input data.

        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        :param Y: additional input snapshots such that Y=AX.
            If not provided, snapshots from X are used.
        :type Y: numpy.ndarray or iterable
        """
        self._reset()
        self._snapshots_holder = Snapshots(X)
        n_samples = self.snapshots.shape[1]

        if Y is None:
            X = self.snapshots[:, :-1]
            Y = self.snapshots[:, 1:]
        else:
            self._compare_data_shapes(Snapshots(Y).snapshots)
            self._snapshots_holder_y = Snapshots(Y)
            X = self.snapshots
            Y = self.snapshots_y

        X, Y = compute_tlsq(X, Y, self._tlsq_rank)
        self._svd_modes, _, _ = self.operator.compute_operator(X, Y)

        # Default timesteps
        self._set_initial_time_dictionary(
            {"t0": 0, "tend": n_samples - 1, "dt": 1}
        )

        self._b = self._compute_amplitudes()

        return self

    def predict(self, X):
        """
        Predict the output Y given the input X using the fitted DMD model.

        :param numpy.ndarray X: the input vector.
        :return: one time-step ahead predicted output.
        :rtype: numpy.ndarray
        """
        return np.linalg.multi_dot(
            [self.modes, np.diag(self.eigs), pinv(self.modes), X]
        )



import numpy as np

from pydmd import SubspaceDMD
from pydmd.subspacedmd import SubspaceDMDOperator, reducedsvd


def test_smoke():
    dmd = SubspaceDMD()
    x = np.random.rand(200, 100)
    assert dmd.fit(x) is not None


def test_modes_shape():
    dmd = SubspaceDMD()
    x = np.random.rand(200, 100)
    assert dmd.fit(x).modes.shape[0] == 200


def test_fixed_parameters():
    dmd = SubspaceDMD()

    assert dmd._tlsq_rank == 0
    assert not dmd.operator._forward_backward
    assert dmd.operator._tikhonov_regularization is None
    assert dmd.operator._exact


def test_default_constructor():
    dmd = SubspaceDMD()

    assert not dmd._opt
    assert dmd.operator._svd_rank == -1
    assert dmd.operator._rescale_mode is None
    assert not dmd.operator._sorted_eigs

    assert dmd._original_time is None
    assert dmd._dmd_time is None
    assert dmd._b is None
    assert dmd._snapshots_holder is None
    assert dmd._modes_activation_bitmask_proxy is None


def test_constructor():
    dmd = SubspaceDMD(
        svd_rank=20,
        opt=True,
        rescale_mode="pippo",
        sorted_eigs="pluto",
    )

    assert dmd._opt
    assert dmd.operator._svd_rank == 20
    assert dmd.operator._rescale_mode == "pippo"
    assert dmd.operator._sorted_eigs == "pluto"


def test_operator():
    dmd = SubspaceDMD()
    assert isinstance(dmd._Atilde, SubspaceDMDOperator)


def test_time():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD().fit(X)

    assert dmd.dmd_time["t0"] == 0
    assert dmd.dmd_time["tend"] == 99
    assert dmd.dmd_time["dt"] == 1


def test_initial_time():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD().fit(X)

    assert dmd.original_time["t0"] == 0
    assert dmd.original_time["tend"] == 99
    assert dmd.original_time["dt"] == 1


def test_amplitudes():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD().fit(X)

    assert dmd.amplitudes is not None


def test_modes():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD().fit(X)

    assert dmd.modes is not None
    assert dmd.modes.shape[0] == 10


def test_eigs():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD().fit(X)

    assert dmd.eigs is not None


def test_snapshots():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD().fit(X)

    assert dmd.snapshots.shape == (10, 100)


def test_reducedsvd():
    X = np.eye(4)
    X = np.hstack((X, X))
    X = np.vstack((X, np.zeros((5, X.shape[1]))))

    U, s, V = reducedsvd(X)
    assert U.shape[1] == 4
    assert V.shape[1] == 4
    assert len(s) == 4


def test_reducedsvd_with_r():
    X = np.eye(4)
    X = np.hstack((X, X))
    X = np.vstack((X, np.zeros((5, X.shape[1]))))

    U, s, V = reducedsvd(X, 2)
    assert U.shape[1] == 2
    assert V.shape[1] == 2
    assert len(s) == 2


def test_svd_rank_positive():
    X = np.random.rand(10, 100)
    dmd = SubspaceDMD(svd_rank=2).fit(X)

    assert len(dmd.eigs) == 2
    assert dmd.modes.shape[1] == 2
    assert len(dmd.amplitudes) == 2



from builtins import range

import numpy as np
from pytest import raises

from pydmd.fbdmd import FbDMD


def create_noisy_data():
    mu = 0.0
    sigma = 0.0  # noise standard deviation
    m = 100  # number of snapshot
    noise = np.random.normal(mu, sigma, m)  # gaussian noise
    A = np.array([[1.0, 1.0], [-1.0, 2.0]])
    A /= np.sqrt(3)
    n = 2
    X = np.zeros((n, m))
    X[:, 0] = np.array([0.5, 1.0])
    # evolve the system and perturb the data with noise
    for k in range(1, m):
        X[:, k] = A.dot(X[:, k - 1])
        X[:, k - 1] += noise[k - 1]
    return X


sample_data = create_noisy_data()


def test_modes_shape():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert dmd.modes.shape[1] == 2


def test_truncation_shape():
    dmd = FbDMD(svd_rank=1)
    dmd.fit(X=sample_data)
    assert dmd.modes.shape[1] == 1


def test_dynamics():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert dmd.dynamics.shape == (2, 100)


def test_eigs_1():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 2


def test_eigs_2():
    dmd = FbDMD(svd_rank=1)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 1


def test_eigs_modulus_1():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    np.testing.assert_almost_equal(np.linalg.norm(dmd.eigs[0]), 1.0, decimal=6)


def test_eigs_modulus_2():
    dmd = FbDMD(svd_rank=-1, exact=True)
    dmd.fit(X=sample_data)
    np.testing.assert_almost_equal(np.linalg.norm(dmd.eigs[1]), 1.0, decimal=6)


def test_reconstructed_data():
    dmd = FbDMD(exact=True, svd_rank=-1)
    dmd.fit(X=sample_data)
    dmd_data = dmd.reconstructed_data
    dmd_data_correct = np.load("tests/test_datasets/fbdmd_data.npy")
    assert np.allclose(dmd_data, dmd_data_correct)


def test_sorted_eigs_default():
    dmd = FbDMD(svd_rank=-1)
    assert dmd.operator._sorted_eigs == False


def test_sorted_eigs_param():
    dmd = FbDMD(svd_rank=-1, sorted_eigs="real")
    assert dmd.operator._sorted_eigs == "real"


def test_get_bitmask_default():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert np.all(dmd.modes_activation_bitmask == True)


def test_set_bitmask():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)

    new_bitmask = np.full(len(dmd.amplitudes), True, dtype=bool)
    new_bitmask[[0]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes_activation_bitmask[0] == False
    assert np.all(dmd.modes_activation_bitmask[1:] == True)


def test_not_fitted_get_bitmask_raises():
    dmd = FbDMD(svd_rank=-1)
    with raises(RuntimeError):
        print(dmd.modes_activation_bitmask)


def test_not_fitted_set_bitmask_raises():
    dmd = FbDMD(svd_rank=-1)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, True, dtype=bool)


def test_raise_wrong_dtype_bitmask():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, 0.1)


def test_fitted():
    dmd = FbDMD(svd_rank=-1)
    assert not dmd.fitted
    dmd.fit(X=sample_data)
    assert dmd.fitted


def test_bitmask_amplitudes():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)

    old_n_amplitudes = dmd.amplitudes.shape[0]
    retained_amplitudes = np.delete(dmd.amplitudes, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.amplitudes.shape[0] == old_n_amplitudes - 2
    np.testing.assert_almost_equal(dmd.amplitudes, retained_amplitudes)


def test_bitmask_eigs():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)

    old_n_eigs = dmd.eigs.shape[0]
    retained_eigs = np.delete(dmd.eigs, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.eigs.shape[0] == old_n_eigs - 2
    np.testing.assert_almost_equal(dmd.eigs, retained_eigs)


def test_bitmask_modes():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)

    old_n_modes = dmd.modes.shape[1]
    retained_modes = np.delete(dmd.modes, [0, -1], axis=1)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes.shape[1] == old_n_modes - 2
    np.testing.assert_almost_equal(dmd.modes, retained_modes)


def test_reconstructed_data_with_bitmask():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=sample_data)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    dmd.reconstructed_data
    assert True


def test_getitem_modes():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=np.load("tests/test_datasets/input_sample.npy"))
    old_n_modes = dmd.modes.shape[1]

    assert dmd[[0, -1]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[0, -1]].modes, dmd.modes[:, [0, -1]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[1::2].modes.shape[1] == old_n_modes // 2
    np.testing.assert_almost_equal(dmd[1::2].modes, dmd.modes[:, 1::2])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[[1, 3]].modes.shape[1] == 2
    np.testing.assert_almost_equal(dmd[[1, 3]].modes, dmd.modes[:, [1, 3]])

    assert dmd.modes.shape[1] == old_n_modes

    assert dmd[2].modes.shape[1] == 1
    np.testing.assert_almost_equal(np.squeeze(dmd[2].modes), dmd.modes[:, 2])

    assert dmd.modes.shape[1] == old_n_modes


def test_getitem_raises():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=np.load("tests/test_datasets/input_sample.npy"))

    with raises(ValueError):
        dmd[[0, 1, 1, 0, 1]]
    with raises(ValueError):
        dmd[[True, True, False, True]]
    with raises(ValueError):
        dmd[1.0]


def test_correct_amplitudes():
    dmd = FbDMD(svd_rank=-1)
    dmd.fit(X=np.load("tests/test_datasets/input_sample.npy"))
    np.testing.assert_array_almost_equal(dmd.amplitudes, dmd._b)



import warnings
import numpy as np
from pytest import raises
from scipy.integrate import solve_ivp
from numpy.testing import assert_equal

from pydmd import DMD
from pydmd import PiDMD
from pydmd import HAVOK

warnings.filterwarnings("ignore")


def generate_lorenz_data(t_eval):
    """
    Given a time vector t_eval = t1, t2, ..., evaluates and returns
    the snapshots of the Lorenz system as columns of the matrix X.
    """

    def lorenz_system(state_t, state):
        sigma, rho, beta = 10, 28, 8 / 3  # chaotic parameters
        x, y, z = state
        x_dot = sigma * (y - x)
        y_dot = (x * (rho - z)) - y
        z_dot = (x * y) - (beta * z)
        return [x_dot, y_dot, z_dot]

    # Set integrator keywords to replicate the odeint defaults
    integrator_keywords = {}
    integrator_keywords["rtol"] = 1e-12
    integrator_keywords["atol"] = 1e-12
    integrator_keywords["method"] = "LSODA"

    sol = solve_ivp(
        lorenz_system,
        [t_eval[0], t_eval[-1]],
        [-8, 8, 27],
        t_eval=t_eval,
        **integrator_keywords,
    )

    return sol.y


# Generate Lorenz system data.
dt = 0.001  # time step
m = 50000  # number of data samples
t = np.arange(m) * dt
X = generate_lorenz_data(t)
x = X[0]


def test_error_fitted():
    """
    Ensure that attempting to get HAVOK attributes results
    in an error if fit() has not yet been called.
    """
    havok = HAVOK()

    with raises(ValueError):
        _ = havok.modes
    with raises(ValueError):
        _ = havok.singular_vals
    with raises(ValueError):
        _ = havok.delay_embeddings
    with raises(ValueError):
        _ = havok.linear_dynamics
    with raises(ValueError):
        _ = havok.forcing
    with raises(ValueError):
        _ = havok.operator
    with raises(ValueError):
        _ = havok.A
    with raises(ValueError):
        _ = havok.B
    with raises(ValueError):
        _ = havok.eigs
    with raises(ValueError):
        _ = havok.r


def test_hankel_1():
    """
    Test that the hankel and dehankel functions work as intended.
    Use 1-D data, lag = 1, and various delay values.
    """
    dummy_data = np.array([1, 2, 3, 4])

    havok = HAVOK(delays=1)
    assert_equal(havok.hankel(dummy_data), np.array([[1, 2, 3, 4]]))
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=2)
    assert_equal(havok.hankel(dummy_data), np.array([[1, 2, 3], [2, 3, 4]]))
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=3)
    assert_equal(havok.hankel(dummy_data), np.array([[1, 2], [2, 3], [3, 4]]))
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=4)
    assert_equal(havok.hankel(dummy_data), np.array([[1, 2, 3, 4]]).T)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    with raises(ValueError):
        havok = HAVOK(delays=5)
        havok.hankel(dummy_data)


def test_hankel_2():
    """
    Test that the hankel and dehankel functions work as intended.
    Use 2-D data, lag = 1, and various delay values.
    """
    dummy_data = np.array([[1, 2, 3], [4, 5, 6]])
    H2 = np.array([[1, 2], [4, 5], [2, 3], [5, 6]])
    H3 = np.array([[1, 4, 2, 5, 3, 6]]).T

    havok = HAVOK(delays=1)
    assert_equal(havok.hankel(dummy_data), dummy_data)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=2)
    assert_equal(havok.hankel(dummy_data), H2)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=3)
    assert_equal(havok.hankel(dummy_data), H3)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    with raises(ValueError):
        havok = HAVOK(delays=4)
        havok.hankel(dummy_data)


def test_hankel_3():
    """
    Test that the hankel and dehankel functions work as intended.
    Use 1-D data, lag = 2, and various delay values.
    """
    dummy_data = np.array([1, 2, 3, 4, 5, 6])
    H1 = np.array([[1, 2, 3, 4, 5, 6]])
    H2 = np.array([[1, 2, 3, 4], [3, 4, 5, 6]])
    H3 = np.array([[1, 2], [3, 4], [5, 6]])

    # If only 1 delay is requested, the lag won't matter.
    havok = HAVOK(delays=1, lag=2)
    assert_equal(havok.hankel(dummy_data), H1)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=2, lag=2)
    assert_equal(havok.hankel(dummy_data), H2)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    havok = HAVOK(delays=3, lag=2)
    assert_equal(havok.hankel(dummy_data), H3)
    assert_equal(havok.dehankel(havok.hankel(dummy_data)), dummy_data)

    with raises(ValueError):
        havok = HAVOK(delays=4, lag=2)
        havok.hankel(dummy_data)


def test_shape_1():
    """
    Using the default HAVOK parameters, checks that the shapes of
    linear_embeddings, forcing_input, A, and B are accurate.
    """
    havok = HAVOK()
    havok.fit(x, t)
    time_length = len(t) - havok.delays + 1
    assert havok.linear_dynamics.shape == (time_length, havok.r - 1)
    assert havok.forcing.shape == (time_length, 1)
    assert havok.A.shape == (havok.r - 1, havok.r - 1)
    assert havok.B.shape == (havok.r - 1, 1)


def test_shape_2():
    """
    Using num_chaos = 2, checks that the shapes of
    linear_embeddings, forcing_input, A, and B are accurate.
    """
    havok = HAVOK(num_chaos=2)
    havok.fit(x, t)
    time_length = len(t) - havok.delays + 1
    assert havok.linear_dynamics.shape == (time_length, havok.r - 2)
    assert havok.forcing.shape == (time_length, 2)
    assert havok.A.shape == (havok.r - 2, havok.r - 2)
    assert havok.B.shape == (havok.r - 2, 2)


def test_snapshots_1():
    """
    Test the stored snapshots and ho_snapshots.
    Ensure that they are accurate in the default case with 1-D data.
    """
    havok = HAVOK()
    with raises(ValueError):
        _ = havok.snapshots
    with raises(ValueError):
        _ = havok.ho_snapshots
    havok.fit(x, t)
    assert_equal(havok.snapshots, x)
    assert_equal(havok.ho_snapshots, havok.hankel(x))


def test_snapshots_2():
    """
    Test the stored snapshots and ho_snapshots.
    Ensure that they are accurate in the default case with 2-D data.
    """
    havok = HAVOK().fit(X, t)
    assert_equal(havok.snapshots, X)
    assert_equal(havok.ho_snapshots, havok.hankel(X))


def test_time_1():
    """
    Test the stored time attribute.
    Ensure that it is accurate in the default case.
    """
    havok = HAVOK()
    with raises(ValueError):
        _ = havok.time
    havok.fit(x, t)
    assert_equal(havok.time, t)


def test_time_2():
    """
    Test that a HAVOK model fitted with a time vector is essentially
    the same as a HAVOK model fitted with the time-step dt. Check the
    stored time vector and the computed HAVOK operator.
    """
    havok_1 = HAVOK().fit(x, t)
    havok_2 = HAVOK().fit(x, dt)
    assert_equal(havok_1.time, havok_2.time)
    assert_equal(havok_1.operator, havok_2.operator)


def test_plot_summary_1():
    """
    Test that everything is fine if we fit a HAVOK
    model and ask for the default summary plot.
    """
    havok = HAVOK(svd_rank=16, delays=100)
    havok.fit(x, t)
    havok.plot_summary()


def test_plot_summary_2():
    """
    Test that everything is fine if we fit a HAVOK model and
    ask for various (but still valid) plot modifications.
    """
    havok = HAVOK(svd_rank=16, delays=100)
    havok.fit(x, t)
    havok.plot_summary(
        num_plot=15000,
        index_linear=(0, 1),
        forcing_threshold=0.005,
        min_jump_dist=200,
        figsize=(15, 4),
        dpi=100,
    )


def test_plot_summary_3():
    """
    Test that an error is thrown if we ask for
    a summary from an unfitted HAVOK model.
    """
    havok = HAVOK()
    with raises(ValueError):
        havok.plot_summary()


def test_reconstruction_1():
    """
    Test the accuracy of the HAVOK reconstruction.
    """
    havok = HAVOK(svd_rank=16, delays=100).fit(x, t)
    error = x - havok.reconstructed_data
    assert np.linalg.norm(error) / np.linalg.norm(x) < 0.4


def test_reconstruction_2():
    """
    Test the accuracy of the sHAVOK reconstruction.
    """
    havok = HAVOK(svd_rank=4, delays=100, structured=True).fit(x, t)
    error = x[:-1] - havok.reconstructed_data
    assert np.linalg.norm(error) / np.linalg.norm(x[:-1]) < 0.4


def test_predict_1():
    """
    Test the accuracy of the HAVOK prediction.
    Ensure that prediction with the HAVOK forcing term and the times
    of fitting simply yields the computed data reconstruction.
    """
    havok = HAVOK(svd_rank=16, delays=100).fit(x, t)
    assert_equal(
        havok.predict(havok.forcing, havok.time[: len(havok.forcing)]),
        havok.reconstructed_data,
    )


def test_predict_2():
    """
    Test the accuracy of the HAVOK prediction.
    Test that predicting beyond the training set isn't absurdly inaccurate.
    """
    havok = HAVOK(svd_rank=16, delays=100).fit(x, t)

    # Build a longer data set and fit a HAVOK model to it.
    t_long = np.arange(2 * m) * dt
    x_long = generate_lorenz_data(t_long)[0]
    havok_long = HAVOK(svd_rank=16, delays=100).fit(x_long, t_long)

    # We only use the long HAVOK model to obtain a long forcing signal.
    forcing_long = havok_long.forcing
    time_long = t_long[: len(forcing_long)]

    # Get the error of the full prediction.
    error = x_long - havok.predict(forcing_long, time_long)
    assert np.linalg.norm(error) / np.linalg.norm(x_long) < 1.0


def test_predict_3():
    """
    Test the accuracy of the HAVOK prediction.
    Test that predicting with V0 indices is functionally
    the same as predicting with array-valued V0 inputs.
    """
    havok = HAVOK(svd_rank=16, delays=100).fit(x, t)
    t_forcing = havok.time[: len(havok.forcing)]
    assert_equal(
        havok.predict(havok.forcing, t_forcing, V0=0),
        havok.predict(havok.forcing, t_forcing, V0=havok.linear_dynamics[0]),
    )
    assert_equal(
        havok.predict(havok.forcing, t_forcing, V0=1),
        havok.predict(havok.forcing, t_forcing, V0=havok.linear_dynamics[1]),
    )
    assert_equal(
        havok.predict(havok.forcing, t_forcing, V0=-1),
        havok.predict(havok.forcing, t_forcing, V0=havok.linear_dynamics[-1]),
    )


def test_threshold_1():
    """
    Test compute_threshold function.
    Test that threshold computation works as expected, whether you plot or not.
    """
    havok = HAVOK(svd_rank=16, delays=100).fit(x, t)
    thres_1 = havok.compute_threshold(p=0.1, bins=100, plot=False)
    thres_2 = havok.compute_threshold(p=0.1, bins=100, plot=True)
    assert thres_1 == thres_2


def test_threshold_2():
    """
    Test compute_threshold function.
    Test that threshold computation works as expected, whether you index
    the stored forcing term, or provide a custom array-valued forcing term.
    """
    havok = HAVOK(svd_rank=16, delays=100).fit(x, t)
    vr = havok.forcing.flatten()
    thres_1 = havok.compute_threshold(p=0.1, bins=100)
    thres_2 = havok.compute_threshold(forcing=vr, p=0.1, bins=100)
    assert thres_1 == thres_2


def test_threshold_3():
    """
    Test compute_threshold function.
    Test that the threshold computation function properly throws an
    error if it is invoked with the default parameters prior to fitting.
    """
    havok = HAVOK(svd_rank=16, delays=100)
    with raises(ValueError):
        _ = havok.compute_threshold()


def test_dmd_1():
    """
    Test that HAVOK works when used with an externally-defined DMD model.
    Test the basic exact DMD model - ensure that plot_summary works just
    fine and that the data reconstruction is still accurate.
    """
    dmd = DMD(svd_rank=-1)
    havok = HAVOK(svd_rank=16, delays=100, dmd=dmd).fit(x, t)
    havok.plot_summary()
    error = x - havok.reconstructed_data
    assert np.linalg.norm(error) / np.linalg.norm(x) < 0.4


def test_dmd_2():
    """
    Test that HAVOK works when used with an externally-defined DMD model.
    Test a physics-informed DMD model - ensure that plot_summary works just
    fine and that the data reconstruction is still accurate.
    """
    pidmd = PiDMD(
        svd_rank=-1,
        manifold="diagonal",
        manifold_opt=2,
        compute_A=True,
    )
    havok = HAVOK(svd_rank=4, delays=100, dmd=pidmd).fit(x, t)
    havok.plot_summary()
    error = x - havok.reconstructed_data
    assert np.linalg.norm(error) / np.linalg.norm(x) < 0.4



"""
Module for DMD plotting.
"""

import warnings
from os.path import splitext

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.patches import Patch

from pydmd import MrDMD

from .bopdmd import BOPDMD
from .hankeldmd import HankelDMD
from .havok import HAVOK
from .preprocessing import PrePostProcessingDMD

mpl.rcParams["figure.max_open_warning"] = 0


def _enforce_ratio(goal_ratio, supx, infx, supy, infy):
    """
    Computes the right value of `supx,infx,supy,infy` to obtain the desired
    ratio in :func:`plot_eigs`. Ratio is defined as
    ::
        dx = supx - infx
        dy = supy - infy
        max(dx,dy) / min(dx,dy)

    :param float goal_ratio: the desired ratio.
    :param float supx: the old value of `supx`, to be adjusted.
    :param float infx: the old value of `infx`, to be adjusted.
    :param float supy: the old value of `supy`, to be adjusted.
    :param float infy: the old value of `infy`, to be adjusted.
    :return tuple: a tuple which contains the updated values of
        `supx,infx,supy,infy` in this order.
    """

    dx = supx - infx
    if dx == 0:
        dx = 1.0e-16
    dy = supy - infy
    if dy == 0:
        dy = 1.0e-16
    ratio = max(dx, dy) / min(dx, dy)

    if ratio >= goal_ratio:
        if dx < dy:
            goal_size = dy / goal_ratio

            supx += (goal_size - dx) / 2
            infx -= (goal_size - dx) / 2
        elif dy < dx:
            goal_size = dx / goal_ratio

            supy += (goal_size - dy) / 2
            infy -= (goal_size - dy) / 2

    return (supx, infx, supy, infy)


def _plot_limits(dmd, narrow_view):
    if narrow_view:
        supx = max(dmd.eigs.real) + 0.05
        infx = min(dmd.eigs.real) - 0.05

        supy = max(dmd.eigs.imag) + 0.05
        infy = min(dmd.eigs.imag) - 0.05

        return _enforce_ratio(8, supx, infx, supy, infy)
    return np.max(np.ceil(np.absolute(dmd.eigs)))


def plot_eigs(
    dmd,
    show_axes=True,
    show_unit_circle=True,
    figsize=(8, 8),
    title="",
    narrow_view=False,
    dpi=None,
    filename=None,
):
    """
    Plot the eigenvalues.

    :param dmd: DMD instance.
    :type dmd: pydmd.DMDBase
    :param bool show_axes: if True, the axes will be showed in the plot.
        Default is True.
    :param bool show_unit_circle: if True, the circle with unitary radius
        and center in the origin will be showed. Default is True.
    :param tuple(int,int) figsize: tuple in inches defining the figure
        size. Default is (8, 8).
    :param str title: title of the plot.
    :param narrow_view bool: if True, the plot will show only the smallest
        rectangular area which contains all the eigenvalues, with a padding
        of 0.05. Not compatible with `show_axes=True`. Default is False.
    :param dpi int: If not None, the given value is passed to
        ``plt.figure``.
    :param str filename: if specified, the plot is saved at `filename`.
    """
    if isinstance(dmd, MrDMD):
        raise ValueError("You should use plot_eigs_mrdmd instead")

    if dmd.eigs is None:
        raise ValueError(
            "The eigenvalues have not been computed."
            "You have to call the fit() method."
        )

    if dpi is not None:
        plt.figure(figsize=figsize, dpi=dpi)
    else:
        plt.figure(figsize=figsize)

    plt.title(title)
    plt.gcf()
    ax = plt.gca()

    points = ax.plot(dmd.eigs.real, dmd.eigs.imag, "bo", label="Eigenvalues")

    if narrow_view:
        supx, infx, supy, infy = _plot_limits(dmd, narrow_view)

        # set limits for axis
        ax.set_xlim((infx, supx))
        ax.set_ylim((infy, supy))

        # x and y axes
        if show_axes:
            endx = np.min([supx, 1.0])
            ax.annotate(
                "",
                xy=(endx, 0.0),
                xytext=(np.max([infx, -1.0]), 0.0),
                arrowprops=dict(arrowstyle=("->" if endx == 1.0 else "-")),
            )

            endy = np.min([supy, 1.0])
            ax.annotate(
                "",
                xy=(0.0, endy),
                xytext=(0.0, np.max([infy, -1.0])),
                arrowprops=dict(arrowstyle=("->" if endy == 1.0 else "-")),
            )
    else:
        # set limits for axis
        limit = _plot_limits(dmd, narrow_view)

        ax.set_xlim((-limit, limit))
        ax.set_ylim((-limit, limit))

        # x and y axes
        if show_axes:
            ax.annotate(
                "",
                xy=(np.max([limit * 0.8, 1.0]), 0.0),
                xytext=(np.min([-limit * 0.8, -1.0]), 0.0),
                arrowprops=dict(arrowstyle="->"),
            )
            ax.annotate(
                "",
                xy=(0.0, np.max([limit * 0.8, 1.0])),
                xytext=(0.0, np.min([-limit * 0.8, -1.0])),
                arrowprops=dict(arrowstyle="->"),
            )

    plt.ylabel("Imaginary part")
    plt.xlabel("Real part")

    if show_unit_circle:
        unit_circle = plt.Circle(
            (0.0, 0.0),
            1.0,
            color="green",
            fill=False,
            label="Unit circle",
            linestyle="--",
        )
        ax.add_artist(unit_circle)

    # Dashed grid
    gridlines = ax.get_xgridlines() + ax.get_ygridlines()
    for line in gridlines:
        line.set_linestyle("-.")
    ax.grid(True)

    # legend
    if show_unit_circle:
        ax.add_artist(
            plt.legend(
                [points, unit_circle],
                ["Eigenvalues", "Unit circle"],
                loc="best",
            )
        )
    else:
        ax.add_artist(plt.legend([points], ["Eigenvalues"], loc="best"))

    ax.set_aspect("equal")

    if filename:
        plt.savefig(filename)
    else:
        plt.show()


def plot_eigs_mrdmd(
    dmd,
    show_axes=True,
    show_unit_circle=True,
    figsize=(8, 8),
    title="",
    level=None,
    node=None,
    filename=None,
):
    """
    Plot the eigenvalues.

    :param bool show_axes: if True, the axes will be showed in the plot.
            Default is True.
    :param bool show_unit_circle: if True, the circle with unitary radius
            and center in the origin will be showed. Default is True.
    :param tuple(int,int) figsize: tuple in inches of the figure.
    :param str title: title of the plot.
    :param int level: plot only the eigenvalues of specific level.
    :param int node: plot only the eigenvalues of specific node.
    :param str filename: if specified, the plot is saved at `filename`.
    """
    if not isinstance(dmd, MrDMD):
        raise ValueError(f"Expected MrDMD, found {type(dmd)}")

    if dmd.eigs is None:
        raise ValueError(
            "The eigenvalues have not been computed."
            "You have to perform the fit method."
        )

    if level:
        peigs = dmd.partial_eigs(level=level, node=node)
    else:
        peigs = dmd.eigs

    plt.figure(figsize=figsize)
    plt.title(title)
    plt.gcf()
    ax = plt.gca()

    if not level:
        cmap = plt.get_cmap("viridis")
        colors = [cmap(i) for i in np.linspace(0, 1, len(dmd.dmd_tree.levels))]

        points = []
        for l in dmd.dmd_tree.levels:
            eigs = dmd.partial_eigs(l)

            points.append(
                ax.plot(eigs.real, eigs.imag, ".", color=colors[l])[0]
            )
    else:
        points = []
        points.append(
            ax.plot(peigs.real, peigs.imag, "bo", label="Eigenvalues")[0]
        )

    # set limits for axis
    limit = np.max(np.ceil(np.absolute(peigs)))
    ax.set_xlim((-limit, limit))
    ax.set_ylim((-limit, limit))

    plt.ylabel("Imaginary part")
    plt.xlabel("Real part")

    if show_unit_circle:
        unit_circle = plt.Circle(
            (0.0, 0.0), 1.0, color="green", fill=False, linestyle="--"
        )
        ax.add_artist(unit_circle)

    # Dashed grid
    gridlines = ax.get_xgridlines() + ax.get_ygridlines()
    for line in gridlines:
        line.set_linestyle("-.")
    ax.grid(True)

    ax.set_aspect("equal")

    # x and y axes
    if show_axes:
        ax.annotate(
            "",
            xy=(np.max([limit * 0.8, 1.0]), 0.0),
            xytext=(np.min([-limit * 0.8, -1.0]), 0.0),
            arrowprops=dict(arrowstyle="->"),
        )
        ax.annotate(
            "",
            xy=(0.0, np.max([limit * 0.8, 1.0])),
            xytext=(0.0, np.min([-limit * 0.8, -1.0])),
            arrowprops=dict(arrowstyle="->"),
        )

    # legend
    if level:
        labels = [f"Eigenvalues - level {level}"]
    else:
        labels = [f"Eigenvalues - level {i}" for i in range(len(points))]

    if show_unit_circle:
        points += [unit_circle]
        labels += ["Unit circle"]

    ax.add_artist(plt.legend(points, labels, loc="best"))
    if filename:
        plt.savefig(filename)
    else:
        plt.show()


def plot_modes_2D(
    dmd,
    snapshots_shape=None,
    index_mode=None,
    filename=None,
    x=None,
    y=None,
    order="C",
    figsize=(8, 8),
):
    """
    Plot the DMD Modes.

    :param dmd: DMD instance.
    :type dmd: pydmd.DMDBase
    :param snapshots_shape: Shape of the snapshots.
    :type tuple: A tuple of ints containing the shape of a single snapshot.
    :param index_mode: the index of the modes to plot. By default, all
        the modes are plotted.
    :type index_mode: int or sequence(int)
    :param str filename: if specified, the plot is saved at `filename`.
    :param numpy.ndarray x: domain abscissa.
    :param numpy.ndarray y: domain ordinate
    :param order: read the elements of snapshots using this index order,
        and place the elements into the reshaped array using this index
        order.  It has to be the same used to store the snapshot. 'C' means
        to read/ write the elements using C-like index order, with the last
        axis index changing fastest, back to the first axis index changing
        slowest.  'F' means to read / write the elements using Fortran-like
        index order, with the first index changing fastest, and the last
        index changing slowest.  Note that the 'C' and 'F' options take no
        account of the memory layout of the underlying array, and only
        refer to the order of indexing.  'A' means to read / write the
        elements in Fortran-like index order if a is Fortran contiguous in
        memory, C-like order otherwise.
    :type order: {'C', 'F', 'A'}, default 'C'.
    :param tuple(int,int) figsize: tuple in inches defining the figure
        size. Default is (8, 8).
    """
    if dmd.modes is None:
        raise ValueError(
            "The modes have not been computed."
            "You have to perform the fit method."
        )

    if snapshots_shape is None:
        snapshots_shape = dmd.snapshots_shape

    if x is None and y is None:
        if snapshots_shape is None:
            raise ValueError(
                "No information about the original shape of the snapshots."
            )

        if len(snapshots_shape) != 2:
            raise ValueError("The dimension of the input snapshots is not 2D.")

    # If domain dimensions have not been passed as argument,
    # use the snapshots dimensions
    if x is None and y is None:
        x = np.arange(snapshots_shape[0])
        y = np.arange(snapshots_shape[1])

    xgrid, ygrid = np.meshgrid(x, y)

    if index_mode is None:
        index_mode = list(range(dmd.modes.shape[1]))
    elif isinstance(index_mode, int):
        index_mode = [index_mode]

    if filename:
        basename, ext = splitext(filename)

    for idx in index_mode:
        fig = plt.figure(figsize=figsize)
        fig.suptitle(f"DMD Mode {idx}")

        real_ax = fig.add_subplot(1, 2, 1)
        imag_ax = fig.add_subplot(1, 2, 2)

        mode = dmd.modes.T[idx].reshape(xgrid.shape, order=order)

        real = real_ax.pcolor(
            xgrid,
            ygrid,
            mode.real,
            cmap="jet",
            vmin=mode.real.min(),
            vmax=mode.real.max(),
        )
        imag = imag_ax.pcolor(
            xgrid,
            ygrid,
            mode.imag,
            vmin=mode.imag.min(),
            vmax=mode.imag.max(),
        )

        fig.colorbar(real, ax=real_ax)
        fig.colorbar(imag, ax=imag_ax)

        real_ax.set_aspect("auto")
        imag_ax.set_aspect("auto")

        real_ax.set_title("Real")
        imag_ax.set_title("Imag")

        # padding between elements
        plt.tight_layout(pad=2.0)

        if filename:
            plt.savefig(f"{basename}.{idx}{ext}")
            plt.close(fig)

    if not filename:
        plt.show()


def plot_snapshots_2D(
    dmd,
    snapshots_shape=None,
    index_snap=None,
    filename=None,
    x=None,
    y=None,
    order="C",
    figsize=(8, 8),
):
    """
    Plot the snapshots.

    :param dmd: DMD instance.
    :type dmd: pydmd.DMDBase
    :param snapshots_shape: Shape of the snapshots.
    :type tuple: A tuple of ints containing the shape of a single snapshot.
    :param index_snap: the index of the snapshots to plot. By default, all
        the snapshots are plotted.
    :type index_snap: int or sequence(int)
    :param str filename: if specified, the plot is saved at `filename`.
    :param numpy.ndarray x: domain abscissa.
    :param numpy.ndarray y: domain ordinate
    :param order: read the elements of snapshots using this index order,
        and place the elements into the reshaped array using this index
        order.  It has to be the same used to store the snapshot. 'C' means
        to read/ write the elements using C-like index order, with the last
        axis index changing fastest, back to the first axis index changing
        slowest.  'F' means to read / write the elements using Fortran-like
        index order, with the first index changing fastest, and the last
        index changing slowest.  Note that the 'C' and 'F' options take no
        account of the memory layout of the underlying array, and only
        refer to the order of indexing.  'A' means to read / write the
        elements in Fortran-like index order if a is Fortran contiguous in
        memory, C-like order otherwise.
    :type order: {'C', 'F', 'A'}, default 'C'.
    :param tuple(int,int) figsize: tuple in inches defining the figure
        size. Default is (8, 8).
    """
    if dmd.snapshots is None:
        raise ValueError("Input snapshots not found.")

    if snapshots_shape is None:
        snapshots_shape = dmd.snapshots_shape

    if x is None and y is None:
        if snapshots_shape is None:
            raise ValueError(
                "No information about the original shape of the snapshots."
            )

        if len(snapshots_shape) != 2:
            raise ValueError("The dimension of the snapshots is not 2D.")

    # If domain dimensions have not been passed as argument,
    # use the snapshots dimensions
    if x is None and y is None:
        x = np.arange(snapshots_shape[0])
        y = np.arange(snapshots_shape[1])

    xgrid, ygrid = np.meshgrid(x, y)

    if index_snap is None:
        index_snap = list(range(dmd.snapshots.shape[1]))
    elif isinstance(index_snap, int):
        index_snap = [index_snap]

    if filename:
        basename, ext = splitext(filename)

    for idx in index_snap:
        fig = plt.figure(figsize=figsize)
        fig.suptitle(f"Snapshot {idx}")

        snapshot = dmd.snapshots.T[idx].real.reshape(xgrid.shape, order=order)

        contour = plt.pcolormesh(
            xgrid,
            ygrid,
            snapshot,
            vmin=snapshot.min(),
            vmax=snapshot.max(),
        )

        fig.colorbar(contour)

        if filename:
            plt.savefig(f"{basename}.{idx}{ext}")
            plt.close(fig)

    if not filename:
        plt.show()


def plot_summary(
    dmd,
    *,
    x=None,
    y=None,
    t=None,
    d=1,
    continuous=False,
    snapshots_shape=None,
    index_modes=(0, 1, 2),
    filename=None,
    order="C",
    figsize=(12, 8),
    dpi=200,
    tight_layout_kwargs=None,
    main_colors=("r", "b", "g"),
    mode_color="k",
    mode_cmap="bwr",
    dynamics_color="tab:blue",
    rank_color="tab:orange",
    circle_color="tab:blue",
    sval_ms=8,
    max_eig_ms=10,
    max_sval_plot=50,
    title_fontsize=14,
    label_fontsize=12,
    plot_semilogy=False,
    flip_continuous_axes=False,
):
    """
    Generate a 3 x 3 summarizing plot that contains the following components:
    - the singular value spectrum of the data
    - the discrete-time and continuous-time DMD eigenvalues
    - the DMD modes specified by the `index_modes` parameter
    - the time dynamics that correspond with each plotted mode
    The number of singular values used for the DMD fit are highlighted.
    All eigenvalues, modes, and dynamics are sorted according to the magnitude
    of their corresponding amplitude value, i.e. their significance in the fit.
    Correspondence between eigenvalues, modes, and dynamics is indicated via
    color coordination.

    :param dmd: fitted DMD instance.
    :type dmd: pydmd.DMDBase
    :param x: Points along the 1st spatial dimension where data has been
        collected.
    :type x: np.ndarray or iterable
    :param y: Points along the 2nd spatial dimension where data has been
        collected. Note that this parameter is only applicable when the data
        snapshots are 2-D, which must be indicated with `snapshots_shape`.
    :type y: np.ndarray or iterable
    :param t: The times of data collection, or the time-step between snapshots.
        Note that time information must be accurate in order to accurately
        visualize eigenvalues and times of the dynamics. For non-`BOPDMD`
        models, the entries of t are assumed to be uniformly-spaced, and if
        not provided, TimeDict information is used. This parameter is ignored
        if an instance of `BOPDMD` is provided.
    :type t: {numpy.ndarray, iterable} or {int, float}
    :param d: Number of delays applied to the data passed to the DMD instance.
        If `d` is greater than 1, then each plotted mode will be the average
        mode taken across all `d` delays.
    :type d: int
    :param continuous: Whether or not the eigenvalues of the given DMD instance
        are continuous-time. If `False`, the eigenvalues are assumed to be the
        discrete-time eigenvalues. If `True`, the eigenvalues are taken to be
        the continuous-time eigenvalues. Note that `continuous` is
        automatically assumed to be `True` if a `BOPDMD` model is given.
    :type continuous: bool
    :param snapshots_shape: Shape of the snapshots. If not provided, the shape
        of the snapshots and modes is assumed to be the flattened space dim of
        the snapshot data. Provide as width, height dimension.
    :type snapshots_shape: iterable
    :param index_modes: Indices of the modes to plot after they have been
        sorted based on significance. At most three may be provided.
        By default, the first three leading modes are plotted.
    :type index_modes: iterable
    :param filename: If specified, the plot is saved at `filename`.
    :type filename: str
    :param order: Read the elements of snapshots using this index order,
        and place the elements into the reshaped array using this index order.
        It has to be the same used to store the snapshot. "C" means to
        read/write the elements using C-like index order, with the last axis
        index changing fastest, back to the first axis index changing slowest.
        "F" means to read/write the elements using Fortran-like index order,
        with the first index changing fastest, and the last index changing
        slowest. Note that the "C" and "F" options take no account of the
        memory layout of the underlying array, and only refer to the order of
        indexing. "A" means to read/write the elements in Fortran-like index
        order if a is Fortran contiguous in memory, C-like order otherwise.
        "C" is used by default.
    :type order: {"C", "F", "A"}
    :param figsize: Width, height in inches.
    :type figsize: iterable
    :param dpi: Figure resolution.
    :type dpi: int
    :param tight_layout_kwargs: Dictionary of `tight_layout` parameters.
    :type tight_layout_kwargs: dict
    :param main_colors: Colors used to denote eigenvalue, mode, dynamics
        associations.
    :type main_colors: iterable
    :param mode_color: Color used to plot the modes, if modes are 1-D.
    :type mode_color: str
    :param mode_cmap: Colormap used to plot the modes, if modes are 2-D.
    :type mode_cmap: str
    :param dynamics_color: Color used to plot the dynamics.
    :type dynamics_color: str
    :param rank_color: Color used to highlight the rank of the DMD fit and
        all DMD eigenvalues aside from those highlighted by `index_modes`.
    :type rank_color: str
    :param circle_color: Color used to plot the unit circle.
    :type circle_color: str
    :param sval_ms: Marker size of all singular values.
    :type sval_ms: int
    :param max_eig_ms: Marker size of the most prominent eigenvalue. The marker
        sizes of all other eigenvalues are then scaled according to eigenvalue
        significance.
    :type max_eig_ms: int
    :param max_sval_plot: Maximum number of singular values to plot.
    :type max_sval_plot: int
    :param title_fontsize: Fontsize used for subplot titles.
    :type title_fontsize: int
    :param label_fontsize: Fontsize used for axis labels.
    :type label_fontsize: int
    :param plot_semilogy: Whether or not to plot the singular values on a
        semilogy plot. If `True`, a semilogy plot is used.
    :type plot_semilogy: bool
    :param flip_continuous_axes: Whether or not to swap the real and imaginary
        axes on the continuous eigenvalues plot. If `True`, the real axis will
        be vertical and the imaginary axis will be horizontal, and vice versa.
    :type flip_continuous_axes: bool
    """

    # This plotting method is inappropriate for plotting HAVOK results.
    if isinstance(dmd, HAVOK):
        raise ValueError("You should use HAVOK.plot_summary() instead.")

    # Check that the DMD instance has been fitted.
    if dmd.modes is None:
        raise ValueError("You need to perform fit() first.")

    # By default, snapshots_shape is the flattened space dimension.
    if snapshots_shape is None:
        snapshots_shape = (len(dmd.snapshots) // d,)
    # If provided, snapshots_shape must contain 2 entires.
    elif len(snapshots_shape) != 2:
        raise ValueError("snapshots_shape must be None or 2-D.")

    # Check the length of index_modes.
    if len(index_modes) > 3:
        raise ValueError("index_modes must have a length of at most 3.")

    # Get the actual rank used for the DMD fit.
    rank = len(dmd.eigs)

    # Ensure that at least rank-many singular values will be plotted.
    if rank > max_sval_plot:
        raise ValueError(f"max_sval_plot must be at least {rank}.")

    # Override index_modes if there are less than 3 modes available.
    if rank < 3:
        warnings.warn(
            "Provided DMD model has less than 3 modes."
            "Plotting all available modes..."
        )
        index_modes = np.arange(rank)

    # Indices cannot go past the total number of available modes.
    if np.any(np.array(index_modes) >= rank):
        raise ValueError(f"Cannot view past mode {rank}.")

    # Sort eigenvalues, modes, and dynamics according to amplitude magnitude.
    mode_order = np.argsort(-np.abs(dmd.amplitudes))
    lead_eigs = dmd.eigs[mode_order]
    lead_modes = dmd.modes[:, mode_order]
    lead_dynamics = dmd.dynamics[mode_order]
    lead_amplitudes = np.abs(dmd.amplitudes[mode_order])

    # Get time information for eigenvalue conversions.
    # The decisions that we make here depend on if we're dealing
    # with a BOPDMD model or any other type of DMD model...
    if isinstance(dmd, BOPDMD) or (
        isinstance(dmd, PrePostProcessingDMD)
        and isinstance(dmd.pre_post_processed_dmd, BOPDMD)
    ):
        # BOPDMD models store time in the time attribute.
        # BOPDMD models also always compute continuous-time eigenvalues.
        cont_eigs = lead_eigs
        time = dmd.time
        dt = dmd.time[1] - dmd.time[0]
        if not np.allclose(dmd.time[1:] - dmd.time[:-1], dt):
            warnings.warn(
                "Time step is not uniform. "
                "No discrete-time eigenvalues to plot..."
            )
            disc_eigs = None
        else:
            disc_eigs = np.exp(lead_eigs * dt)
    else:
        # For all other dmd models, go to the TimeDict for time information,
        # that or use the user-provided time information in t if available.
        num_samples = dmd.snapshots.shape[-1]
        if isinstance(t, (int, float)):
            time = np.arange(num_samples) * t
            dt = t
        elif t is not None:
            time = np.squeeze(np.array(t))
            dt = time[1] - time[0]
            if not np.allclose(time[1:] - time[:-1], dt):
                warnings.warn(
                    "Time step is not uniform. DMD might produce unexpected "
                    "results. Consider using BOP-DMD instead."
                )
        else:
            try:
                time = dmd.original_timesteps
                dt = dmd.original_time["dt"]
            except AttributeError:
                warnings.warn(
                    "No time information available. Using dt = 1 and t0 = 0."
                )
                time = np.arange(num_samples)
                dt = 1.0

        if continuous:
            cont_eigs = lead_eigs
            disc_eigs = np.exp(lead_eigs * dt)
        else:
            disc_eigs = lead_eigs
            cont_eigs = np.log(lead_eigs) / dt

    # Get mode averages across delays if time-delay was used.
    if d > 1:
        lead_modes = np.average(
            lead_modes.reshape(
                d,
                lead_modes.shape[0] // d,
                lead_modes.shape[1],
            ),
            axis=0,
        )

    # Compute the singular values of the data matrix.
    if isinstance(dmd, HankelDMD):
        # Use time-delay data matrix to compute singular values.
        snp = dmd.ho_snapshots
    else:
        # Use input data matrix to compute singular values.
        snp = dmd.snapshots
    s = np.linalg.svd(snp, full_matrices=False, compute_uv=False)
    # Compute the percent of data variance captured by each singular value.
    s_var = s * (100 / np.sum(s))
    s_var = s_var[:max_sval_plot]

    # Build a list of indices of the complex conjugate pairs to highlight.
    # Example: If index_modes = [idx1, idx2, idx3, idx4], such that...
    #   idx1 has no complex conjugate pair
    #   idx2 and idx3 are complex conjugates
    #   idx4 and idx5 are complex conjugates
    # Then index_modes_cc = [(idx1, idx1), (idx2, idx3), (idx4, idx5)]
    index_modes_cc = []
    for idx in index_modes:
        eig = cont_eigs[idx]
        if eig.conj() not in cont_eigs:
            index_modes_cc.append((idx, idx))
        elif idx not in np.array(index_modes_cc):
            index_modes_cc.append((idx, list(cont_eigs).index(eig.conj())))
    other_eigs = np.setdiff1d(np.arange(rank), np.array(index_modes_cc))

    # Generate the summarizing plot.
    fig, (eig_axes, mode_axes, dynamics_axes) = plt.subplots(
        3,
        3,
        figsize=figsize,
        dpi=dpi,
    )

    # PLOT 1: Plot the singular value spectrum.
    eig_axes[0].set_title("Singular Values", fontsize=title_fontsize)
    eig_axes[0].set_ylabel("% variance", fontsize=label_fontsize)
    s_t = np.arange(len(s_var)) + 1
    eig_axes[0].plot(
        s_t[:rank],
        s_var[:rank],
        "o",
        c=rank_color,
        ms=sval_ms,
        mec="k",
    )
    eig_axes[0].plot(
        s_t[rank:],
        s_var[rank:],
        "o",
        c="gray",
        ms=sval_ms,
        mec="k",
    )
    eig_axes[0].legend(
        handles=[Patch(facecolor=rank_color, label="Rank of fit")]
    )
    if plot_semilogy:
        eig_axes[0].semilogy()

    # PLOTS 2-3: Plot the eigenvalues (discrete-time and continuous-time).
    # Scale marker sizes to reflect their associated amplitude.
    ms_vals = max_eig_ms * np.sqrt(lead_amplitudes / lead_amplitudes[0])

    # PLOT 2: Plot the discrete-time eigenvalues on the unit circle.
    eig_axes[1].axvline(x=0, c="k", lw=1)
    eig_axes[1].axhline(y=0, c="k", lw=1)
    eig_axes[1].axis("equal")
    eig_axes[1].set_title("Discrete-time Eigenvalues", fontsize=title_fontsize)
    t = np.linspace(0, 2 * np.pi, 100)
    eig_axes[1].plot(np.cos(t), np.sin(t), c=circle_color, ls="--")
    eig_axes[1].set_xlabel(r"$Re(\lambda)$", fontsize=label_fontsize)
    eig_axes[1].set_ylabel(r"$Im(\lambda)$", fontsize=label_fontsize)

    # PLOT 3: Plot the continuous-time eigenvalues.
    eig_axes[2].axvline(x=0, c="k", lw=1)
    eig_axes[2].axhline(y=0, c="k", lw=1)
    eig_axes[2].axis("equal")
    eig_axes[2].set_title(
        "Continuous-time Eigenvalues",
        fontsize=title_fontsize,
    )
    if flip_continuous_axes:
        eig_axes[2].set_xlabel(r"$Im(\omega)$", fontsize=label_fontsize)
        eig_axes[2].set_ylabel(r"$Re(\omega)$", fontsize=label_fontsize)
        eig_axes[2].invert_xaxis()
        cont_eigs = 1j * cont_eigs.real + cont_eigs.imag
    else:
        eig_axes[2].set_xlabel(r"$Re(\omega)$", fontsize=label_fontsize)
        eig_axes[2].set_ylabel(r"$Im(\omega)$", fontsize=label_fontsize)

    # Now plot the eigenvalues and record the colors used for each main index.
    mode_colors = {}
    for ax, eigs in zip([eig_axes[1], eig_axes[2]], [disc_eigs, cont_eigs]):
        if eigs is not None:
            # Plot the main indices and their complex conjugate.
            for i, indices in enumerate(index_modes_cc):
                for idx in indices:
                    ax.plot(
                        eigs[idx].real,
                        eigs[idx].imag,
                        "o",
                        c=main_colors[i],
                        ms=ms_vals[idx],
                        mec="k",
                    )
                    mode_colors[idx] = main_colors[i]
            # Plot all other DMD eigenvalues.
            for idx in other_eigs:
                ax.plot(
                    eigs[idx].real,
                    eigs[idx].imag,
                    "o",
                    c=rank_color,
                    ms=ms_vals[idx],
                    mec="k",
                )

    # Build the spatial grid for the mode plots.
    if x is None:
        x = np.arange(snapshots_shape[0])
    if len(snapshots_shape) == 2:
        if y is None:
            y = np.arange(snapshots_shape[1])
        xgrid, ygrid = np.meshgrid(x, y)

    # PLOTS 4-6: Plot the DMD modes.
    for i, (ax, idx) in enumerate(zip(mode_axes, index_modes)):
        ax.set_title(
            f"Mode {idx + 1}",
            c=mode_colors[idx],
            fontsize=title_fontsize,
        )
        if len(snapshots_shape) == 1:
            # Plot modes in 1-D.
            ax.plot(x, lead_modes[:, idx].real, c=mode_color)
        else:
            # Plot modes in 2-D.
            mode = lead_modes[:, idx].reshape(xgrid.shape, order=order)
            vmax = np.abs(mode.real).max()
            im = ax.pcolormesh(
                xgrid,
                ygrid,
                mode.real,
                vmax=vmax,
                vmin=-vmax,
                cmap=mode_cmap,
            )
            # Align the colorbar with the plotted image.
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="3%", pad=0.05)
            fig.colorbar(im, cax=cax)

    # PLOTS 7-9: Plot the DMD mode dynamics.
    for i, (ax, idx) in enumerate(zip(dynamics_axes, index_modes)):
        dynamics_data = lead_dynamics[idx].real
        ax.set_title(
            "Mode Dynamics",
            c=mode_colors[idx],
            fontsize=title_fontsize,
        )
        ax.plot(time, dynamics_data, c=dynamics_color)
        ax.set_xlabel("Time", fontsize=label_fontsize)

        # Re-adjust ylim if dynamics oscillations are extremely small.
        dynamics_range = dynamics_data.max() - dynamics_data.min()
        if dynamics_range / np.abs(np.average(dynamics_data)) < 1e-4:
            ax.set_ylim(np.sort([0.0, 2 * np.average(dynamics_data)]))

    # Padding between elements.
    if tight_layout_kwargs is None:
        tight_layout_kwargs = {}
    plt.tight_layout(**tight_layout_kwargs)

    # Save plot if filename is provided.
    if filename:
        plt.savefig(filename)
        plt.close(fig)
    else:
        plt.show()
